import unittest
import numpy as np
from scipy.stats import norm
import six
from six import StringIO
import abc
import os

from kafe2.fit.representation import FitYamlWriter, FitYamlReader
from kafe2.fit.io.handle import IOStreamHandle
from kafe2.fit.custom import CustomFit
from kafe2.fit.histogram import HistFit
from kafe2.fit.indexed import IndexedFit
from kafe2.fit.unbinned import UnbinnedFit
from kafe2.fit.xy import XYFit
from kafe2.fit.histogram.container import HistContainer
from .test_fit_custom import TestCustomFitWithSimpleYErrors


@six.add_metaclass(abc.ABCMeta)
class AbstractTestFitRepresenter(object):
    FIT_CLASS = None
    TEST_FIT = None
    TEST_FIT_SIMPLE = None
    TEST_FIT_MISSING_KEYWORD = None
    TEST_FIT_EXTRA_KEYWORD = None

    @abc.abstractmethod
    def setUp(self):
        pass

    def tearDown(self):
        os.remove(self._state_filename)

    def setup_streams(self):
        self._testfile_stringstream = IOStreamHandle(StringIO(self.TEST_FIT))
        self._testfile_stringstream_simple = IOStreamHandle(StringIO(self.TEST_FIT_SIMPLE))
        self._testfile_streamreader = FitYamlReader(self._testfile_stringstream)
        self._testfile_streamreader_simple = FitYamlReader(self._testfile_stringstream_simple)

        self._roundtrip_stringstream = IOStreamHandle(StringIO())
        self._state_filename = f"test-{self.FIT_CLASS.__name__}.yml"
        self._roundtrip_streamreader = FitYamlReader(self._roundtrip_stringstream)
        self._roundtrip_streamwriter = FitYamlWriter(self._fit, self._roundtrip_stringstream)

        self._testfile_stringstream_missing_keyword = IOStreamHandle(StringIO(self.TEST_FIT_MISSING_KEYWORD))
        self._testfile_stringstream_extra_keyword = IOStreamHandle(StringIO(self.TEST_FIT_EXTRA_KEYWORD))
        self._testfile_streamreader_missing_keyword = FitYamlReader(self._testfile_stringstream_missing_keyword)
        self._testfile_streamreader_extra_keyword = FitYamlReader(self._testfile_stringstream_extra_keyword)

    def test_write_to_roundtrip_stringstream(self):
        self._roundtrip_streamwriter.write()

    def test_read_from_testfile_stream_missing_keyword(self):
        with self.assertRaises(ValueError):
            self._testfile_streamreader_missing_keyword.read()

    def test_read_from_testfile_stream_extra_keyword(self):
        with self.assertRaises(ValueError):
            self._testfile_streamreader_extra_keyword.read()

    def test_round_trip_save_results(self):
        self._fit.do_fit()
        self._roundtrip_streamwriter.write()
        self._roundtrip_stringstream.seek(0)  # return to beginning
        _read_fit = self._roundtrip_streamreader.read()
        self.assertTrue(isinstance(_read_fit, self.FIT_CLASS))

        self.assertTrue(self._fit.did_fit == _read_fit.did_fit)
        self.assertTrue(
            np.allclose(
                self._fit.parameter_cov_mat,
                _read_fit.parameter_cov_mat
            )
        )
        self.assertTrue(
            np.allclose(
                self._fit.parameter_errors,
                _read_fit.parameter_errors
            )
        )
        self.assertTrue(
            np.allclose(
                self._fit.parameter_cor_mat,
                _read_fit.parameter_cor_mat
            )
        )

    def test_round_trip_save_results_2(self):
        self._roundtrip_streamwriter.write()
        self._roundtrip_stringstream.seek(0)  # return to beginning
        _read_fit = self._roundtrip_streamreader.read()

        self._fit.do_fit()
        self._fit.save_state(self._state_filename)
        _read_fit.load_state(self._state_filename)

        self.assertTrue(isinstance(_read_fit, self.FIT_CLASS))

        self.assertTrue(self._fit.did_fit == _read_fit.did_fit)
        self.assertTrue(
            np.allclose(
                self._fit.parameter_cov_mat,
                _read_fit.parameter_cov_mat
            )
        )
        self.assertTrue(
            np.allclose(
                self._fit.parameter_errors,
                _read_fit.parameter_errors
            )
        )
        self.assertTrue(
            np.allclose(
                self._fit.parameter_cor_mat,
                _read_fit.parameter_cor_mat
            )
        )


TEST_FIT_CUSTOM="""
type: custom
cost_function: |
    def chi2(a=1.5, b=-0.5):
        x_data = np.array([0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0])
        y_data = np.array([
            -1.0804945, 0.97336504, 2.75769933, 4.91093935, 6.98511206, 9.15059627, 10.9665515,
            13.06741151, 14.95081026, 16.94404467
        ])
        y_err = 0.1
        y_model = a * x_data + b
        cost = np.sum(np.square((y_model - y_data) / y_err))
        cost += 2 * x_data.shape[0] * np.log(y_err)
        return cost
parameter_constraints:
      - type: simple
        index: 0
        value: 2.0
        uncertainty: 1.0
      - type: simple
        name: b
        value: -1.0
        uncertainty: 0.5
      - type: matrix
        names: [a, b]
        values: [2.05, -0.95]
        matrix: [[1.1, 0.1], [0.1, 2.4]]
parameter_formatters:
    a: alpha
    b: beta
"""

TEST_FIT_CUSTOM_MISSING_KEYWORD="""
type: custom
"""

TEST_FIT_CUSTOM_EXTRA_KEYWORD = TEST_FIT_CUSTOM + """
extra_keyword: 3.14
"""

class TestCustomFitYamlRepresenter(unittest.TestCase, AbstractTestFitRepresenter):
    FIT_CLASS = CustomFit
    TEST_FIT = TEST_FIT_CUSTOM
    TEST_FIT_SIMPLE = None
    TEST_FIT_EXTRA_KEYWORD = TEST_FIT_CUSTOM_EXTRA_KEYWORD
    TEST_FIT_MISSING_KEYWORD = TEST_FIT_CUSTOM_MISSING_KEYWORD

    def setUp(self):
        self._test_parameters = np.array([2.0, -1.0])
        self._test_parameters_default = np.array([1.5, -0.5])
        self._test_parameters_do_fit = np.array([2.0115580399995032, -1.0889949779758534])
        self._test_y_default = self._test_parameters_default[0] * np.arange(10) \
                               + self._test_parameters_default[1]
        self._test_y_do_fit = [-1.08899498, 0.92256306, 2.9341211, 4.94567914, 6.95723718,
                               8.96879522, 10.98035326,
                               12.9919113, 15.00346934, 17.01502738]

        self._fit = CustomFit(TestCustomFitWithSimpleYErrors.chi2)
        self._fit.assign_parameter_names(a='alpha', b='beta')
        self._fit.add_parameter_constraint(name='a', value=2.0, uncertainty=1.0)
        self._fit.add_parameter_constraint(name='b', value=-1.0, uncertainty=0.5)
        self._fit.add_matrix_parameter_constraint(names=['a', 'b'], values=[2.05, -0.95],
                                                  matrix=[[1.1, 0.1], [0.1, 2.4]])
        self.setup_streams()

    def test_read_from_testfile_stream(self):
        _read_fit = self._testfile_streamreader.read()
        self.assertTrue(isinstance(_read_fit, self.FIT_CLASS))
        for _pf in _read_fit._parameter_formatters:
            self.assertTrue(
                _pf.arg_name == 'a' and _pf.name == 'alpha'
                or _pf.arg_name == 'b' and _pf.name == 'beta'
            )
        self.assertTrue(
            np.allclose(
                self._test_parameters_default,
                _read_fit.parameter_values
            )
        )

        _read_fit.do_fit()

        self.assertTrue(
            np.allclose(
                self._test_parameters_do_fit,
                _read_fit.parameter_values
            )
        )

    def test_round_trip_with_stringstream(self):
        self._roundtrip_streamwriter.write()
        self._roundtrip_stringstream.seek(0)  # return to beginning
        _read_fit = self._roundtrip_streamreader.read()
        self.assertTrue(isinstance(_read_fit, self.FIT_CLASS))
        for _pf in _read_fit._parameter_formatters:
            self.assertTrue(
                _pf.arg_name == 'a' and _pf.name == 'alpha'
                or _pf.arg_name == 'b' and _pf.name == 'beta'
            )

        self.assertTrue(
            np.allclose(
                self._test_parameters_default,
                _read_fit.parameter_values
            )
        )

        _read_fit.do_fit()

        self.assertTrue(
            np.allclose(
                self._test_parameters_do_fit,
                _read_fit.parameter_values
            )
        )


TEST_FIT_HIST="""
type: histogram
cost_function: nll_poisson
dataset:
    type: histogram
    n_bins: 8
    bin_range: [-2.0, 2.0]
    raw_data: [ 0.94657578,  0.3674123,  -0.16462338,  0.0408205,   0.35318031, -0.35919436,
               -0.22111037, -0.4663712,  -0.26321687,  0.63243935,  0.30995581,  0.45732282,
                0.89150917,  0.62005489, -0.5260296,   0.96317915,  0.2918086,   0.27343654,
                0.0070309,  -0.54261172,  0.01499711,  0.32934897,  0.53675086, -0.33745179,
                0.42785108,  0.29124784,  0.82751196,  0.3073723,  -0.51430445, -0.5821153,
                0.28015453, -0.06767495, -0.42469716,  0.08997751, -0.34496342,  0.91208844,
                0.48666838,  0.54641629, -0.26780562, -0.18634414, -0.87815388,  0.01707615,
                0.62112113,  1.2714954,   1.00311742, -0.65583951, -0.51491734, -0.13001327,
                0.12267431,  0.06186063, -0.25203,     0.5991445,  -0.05863787,  0.44575788,
               -0.55761754,  0.70046957,  1.0686471,   0.04027402,  0.39745425,  0.8821956,
                0.2241812,   0.01701976, -0.44040885,  0.87585192,  0.62350494,  0.34860764,
               -0.53451919,  0.7347968,   0.32077684,  1.0450513,   0.467221,    0.46268865,
                0.66497412, -0.48091556,  0.33789605,  0.46706738,  0.23336683,  0.29909525,
                0.08691923,  0.98999353, -0.17895406,  0.50533812,  0.81767467, -0.76434111,
                0.09429677,  0.61596385,  0.69613775, -0.29647721,  0.48768034, -0.44852791]
    errors:
      - correlation_coefficient: 0.0
        error_value: 0.1
        name: test_error
        relative: false
        type: simple
parametric_model:
    type: histogram
    n_bins: 8
    bin_range: [-2.0, 2.0]
    model_density_function:
        type: histogram
        python_code: 'x mu sigma -> exp(-0.5 * ((x - mu) / sigma)^2) / sqrt(2 * pi * sigma ** 2)'
    model_parameters: [0.1, 1.0]
parameter_constraints:
      - type: simple
        index: 0
        value: 0.1
        uncertainty: 1.0
      - type: simple
        name: sigma
        value: 1.0
        uncertainty: 0.5
      - type: matrix
        names: [sigma, mu]
        values: [1.3, 2.5]
        matrix: [[1.1, 0.1], [0.1, 2.4]]
"""

TEST_FIT_HIST_MISSING_KEYWORD="""
type: histogram
parametric_model:
    type: histogram
    n_bins: 8
    bin_range: [-2.0, 2.0]
    model_density_function:
        type: histogram
        python_code: |
            def hist_model_density(x, mu, sigma):
                return np.exp(-0.5 * ((x - mu) / sigma) ** 2) / np.sqrt(2 * np.pi * sigma ** 2)
    model_parameters: [0.1, 1.0]
"""

TEST_FIT_HIST_EXTRA_KEYWORD = TEST_FIT_HIST + """
extra_keyword: 3.14
"""

TEST_FIT_HIST_SIMPLE="""
type: histogram
n_bins: 8
bin_range: [-2.0, 2.0]
raw_data: [ 0.94657578,  0.3674123,  -0.16462338,  0.0408205,   0.35318031, -0.35919436,
           -0.22111037, -0.4663712,  -0.26321687,  0.63243935,  0.30995581,  0.45732282,
            0.89150917,  0.62005489, -0.5260296,   0.96317915,  0.2918086,   0.27343654,
            0.0070309,  -0.54261172,  0.01499711,  0.32934897,  0.53675086, -0.33745179,
            0.42785108,  0.29124784,  0.82751196,  0.3073723,  -0.51430445, -0.5821153,
            0.28015453, -0.06767495, -0.42469716,  0.08997751, -0.34496342,  0.91208844,
            0.48666838,  0.54641629, -0.26780562, -0.18634414, -0.87815388,  0.01707615,
            0.62112113,  1.2714954,   1.00311742, -0.65583951, -0.51491734, -0.13001327,
            0.12267431,  0.06186063, -0.25203,     0.5991445,  -0.05863787,  0.44575788,
           -0.55761754,  0.70046957,  1.0686471,   0.04027402,  0.39745425,  0.8821956,
            0.2241812,   0.01701976, -0.44040885,  0.87585192,  0.62350494,  0.34860764,
           -0.53451919,  0.7347968,   0.32077684,  1.0450513,   0.467221,    0.46268865,
            0.66497412, -0.48091556,  0.33789605,  0.46706738,  0.23336683,  0.29909525,
            0.08691923,  0.98999353, -0.17895406,  0.50533812,  0.81767467, -0.76434111,
            0.09429677,  0.61596385,  0.69613775, -0.29647721,  0.48768034, -0.44852791]
errors: 0.1
"""


class TestHistFitYamlRepresenter(unittest.TestCase, AbstractTestFitRepresenter):
    FIT_CLASS = HistFit
    TEST_FIT = TEST_FIT_HIST
    TEST_FIT_SIMPLE = TEST_FIT_HIST_SIMPLE
    TEST_FIT_EXTRA_KEYWORD = TEST_FIT_HIST_EXTRA_KEYWORD
    TEST_FIT_MISSING_KEYWORD = TEST_FIT_HIST_MISSING_KEYWORD

    @staticmethod
    def hist_model_density(x, mu=0.1, sigma=1.0):
        #TODO scipy.stats.norm support
        return np.exp(-0.5 * ((x - mu) / sigma) ** 2) / np.sqrt(2 * np.pi * sigma ** 2)
    
    @staticmethod
    def hist_model_density_antiderivative(x, mu, sigma):
        return norm(mu, sigma).cdf(x)
    
    def setUp(self):
        self._test_raw_data = [
            0.94657578,  0.3674123,  -0.16462338,  0.0408205,   0.35318031, -0.35919436,
           -0.22111037, -0.4663712,  -0.26321687,  0.63243935,  0.30995581,  0.45732282,
            0.89150917,  0.62005489, -0.5260296,   0.96317915,  0.2918086,   0.27343654,
            0.0070309,  -0.54261172,  0.01499711,  0.32934897,  0.53675086, -0.33745179,
            0.42785108,  0.29124784,  0.82751196,  0.3073723,  -0.51430445, -0.5821153,
            0.28015453, -0.06767495, -0.42469716,  0.08997751, -0.34496342,  0.91208844,
            0.48666838,  0.54641629, -0.26780562, -0.18634414, -0.87815388,  0.01707615,
            0.62112113,  1.2714954,   1.00311742, -0.65583951, -0.51491734, -0.13001327,
            0.12267431,  0.06186063, -0.25203,     0.5991445,  -0.05863787,  0.44575788,
           -0.55761754,  0.70046957,  1.0686471,   0.04027402,  0.39745425,  0.8821956,
            0.2241812,   0.01701976, -0.44040885,  0.87585192,  0.62350494,  0.34860764,
           -0.53451919,  0.7347968,   0.32077684,  1.0450513,   0.467221,    0.46268865,
            0.66497412, -0.48091556,  0.33789605,  0.46706738,  0.23336683,  0.29909525,
            0.08691923,  0.98999353, -0.17895406,  0.50533812,  0.81767467, -0.76434111,
            0.09429677,  0.61596385,  0.69613775, -0.29647721,  0.48768034, -0.44852791
        ]
        self._test_n_bins = 8
        self._test_parameters_default = np.array([0.1, 1.0])
        self._test_parameters_default_simple = np.array([1.0, 1.0])
        self._test_parameters = np.array([0.2, 0.5])
        self._test_bin_range = (-2.0, 2.0)
        self._test_parameters_do_fit = np.array([0.20292002, 0.50108339])
        self._test_parameters_do_fit_simple = np.array([0.20032218, 0.49665461])
        self._test_x = np.array([-2.0, -1.5, -1.0, -0.5, 0.0, 0.5, 1.0, 1.5, 2.0])
        self._test_density_default = self.hist_model_density(self._test_x, *self._test_parameters_default)
        self._test_density_default_simple = self.hist_model_density(self._test_x, *self._test_parameters_default_simple)
        self._test_density_do_fit = self.hist_model_density(self._test_x, *self._test_parameters_do_fit)
        self._test_density_do_fit_simple = self.hist_model_density(self._test_x, *self._test_parameters_do_fit_simple)

        _data = HistContainer(n_bins=self._test_n_bins, bin_range=self._test_bin_range, fill_data=self._test_raw_data)
        self._fit = HistFit(
            data=_data,
            model_function=TestHistFitYamlRepresenter.hist_model_density
        )
        self._fit.add_error(err_val=0.1)
        self._fit.add_parameter_constraint("mu", 0.1, 1.0)
        self._fit.add_parameter_constraint("sigma", 1.0, 0.5)
        self._fit.add_matrix_parameter_constraint(
            ["sigma", "mu"], [1.3, 2.5], [[1.1, 0.1], [0.1, 2.4]])

        self.setup_streams()

    def test_read_from_testfile_stream(self):
        _read_fit = self._testfile_streamreader.read()
        self.assertTrue(isinstance(_read_fit, self.FIT_CLASS))
        self.assertTrue(
            np.allclose(
                self._test_parameters_default,
                _read_fit.parameter_values
            )
        )
        self.assertTrue(
            np.allclose(
                self._test_density_default,
                _read_fit.eval_model_function_density(self._test_x)
            )
        )
        
        _read_fit.do_fit()
        self.assertTrue(
            np.allclose(
                self._test_parameters_do_fit,
                _read_fit.parameter_values
            )
        )
        self.assertTrue(
            np.allclose(
                self._test_density_do_fit,
                _read_fit.eval_model_function_density(self._test_x)
            )
        )

    def test_read_from_testfile_stream_simple(self):
        _read_fit = self._testfile_streamreader_simple.read()
        self.assertTrue(isinstance(_read_fit, self.FIT_CLASS))
        self.assertTrue(
            np.allclose(
                self._test_parameters_default_simple,
                _read_fit.parameter_values
            )
        )
        self.assertTrue(
            np.allclose(
                self._test_density_default_simple,
                _read_fit.eval_model_function_density(self._test_x)
            )
        )
        
        _read_fit.do_fit()

        self.assertTrue(
            np.allclose(
                self._test_parameters_do_fit_simple,
                _read_fit.parameter_values
            )
        )
        self.assertTrue(
            np.allclose(
                self._test_density_do_fit_simple,
                _read_fit.eval_model_function_density(self._test_x)
            )
        )

    def test_round_trip_with_stringstream(self):
        self._roundtrip_streamwriter.write()
        self._roundtrip_stringstream.seek(0)  # return to beginning
        _read_fit = self._roundtrip_streamreader.read()
        self.assertTrue(isinstance(_read_fit, self.FIT_CLASS))

        self.assertTrue(
            np.allclose(
                self._test_parameters_default,
                _read_fit.parameter_values
            )
        )
        self.assertTrue(
            np.allclose(
                self._test_density_default,
                _read_fit.eval_model_function_density(self._test_x)
            )
        )
        
        _read_fit.do_fit()

        self.assertTrue(
            np.allclose(
                self._test_parameters_do_fit,
                _read_fit.parameter_values
            )
        )
        self.assertTrue(
            np.allclose(
                self._test_density_do_fit,
                _read_fit.eval_model_function_density(self._test_x)
            )
        )


TEST_FIT_INDEXED="""
type: indexed
cost_function: chi2
dataset:
    type: indexed
    data: [ -1.0804945, 0.97336504, 2.75769933, 4.91093935, 6.98511206,
           9.15059627, 10.9665515, 13.06741151, 14.95081026, 16.94404467]
    errors:
      - correlation_coefficient: 0.0
        error_value: 0.1
        name: test_error
        relative: false
        type: simple
parametric_model:
    type: indexed
    model_function:
        type: indexed
        python_code: |
            def linear_model(a, b):
                return a * np.arange(10) + b
    model_parameters: [1.5, -0.5]
parameter_constraints:
      - type: simple
        index: 0
        value: 2.0
        uncertainty: 1.0
      - type: simple
        name: b
        value: -1.0
        uncertainty: 0.5
      - type: matrix
        names: [a, b]
        values: [2.05, -0.95]
        matrix: [[1.1, 0.1], [0.1, 2.4]]
"""

TEST_FIT_INDEXED_MISSING_KEYWORD="""
type: indexed
parametric_model:
    type: indexed
    model_function:
        type: indexed
        python_code: |
            def linear_model(a, b):
                return a * np.arange(10) + b
    model_parameters: [1.5, -0.5]
"""

TEST_FIT_INDEXED_EXTRA_KEYWORD = TEST_FIT_INDEXED + """
extra_keyword: 3.14
"""

TEST_FIT_INDEXED_SIMPLE="""
type: indexed
data: [ -1.0804945, 0.97336504, 2.75769933, 4.91093935, 6.98511206,
       9.15059627, 10.9665515, 13.06741151, 14.95081026, 16.94404467]
errors: 0.1
model_function: |
    def linear_model(a, b):
        return a * np.arange(10) + b
"""


class TestIndexedFitYamlRepresenter(unittest.TestCase, AbstractTestFitRepresenter):
    FIT_CLASS = IndexedFit
    TEST_FIT = TEST_FIT_INDEXED
    TEST_FIT_SIMPLE = TEST_FIT_INDEXED_SIMPLE
    TEST_FIT_EXTRA_KEYWORD = TEST_FIT_INDEXED_EXTRA_KEYWORD
    TEST_FIT_MISSING_KEYWORD = TEST_FIT_INDEXED_MISSING_KEYWORD

    @staticmethod
    def linear_model(a=1.5, b=-0.5):
        return a * np.arange(10) + b
    
    def setUp(self):
        self._test_x = np.arange(10)
        self._test_parameters = np.array([2.0, -1.0])
        self._test_parameters_default = np.array([1.5, -0.5])
        self._test_parameters_default_simple = np.array([1.0, 1.0])
        self._test_parameters_do_fit = np.array([2.0115580399995032, -1.0889949779758534])
        self._test_parameters_do_fit_simple = np.array([2.0117809129092095, -1.090410559090481])
        self._test_y = [-1.0804945, 0.97336504, 2.75769933, 4.91093935, 6.98511206,
                        9.15059627, 10.9665515, 13.06741151, 14.95081026, 16.94404467]
        self._test_y_default = self.linear_model(*self._test_parameters_default)
        self._test_y_default_simple = self.linear_model(*self._test_parameters_default_simple)
        self._test_y_do_fit = [-1.08899498, 0.92256306, 2.9341211, 4.94567914, 6.95723718, 8.96879522,
                               10.98035326, 12.9919113, 15.00346934, 17.01502738]
        self._test_y_do_fit_simple = [-1.09041056, 0.92137035, 2.93315127, 4.94493218, 6.95671309,
                                      8.96849401, 10.98027492, 12.99205583, 15.00383674, 17.01561766]

        self._fit = IndexedFit(
            data=self._test_y,
            model_function=TestIndexedFitYamlRepresenter.linear_model
        )
        self._fit.set_all_parameter_values(self._test_parameters_default)
        self._fit.add_error(err_val=0.1)
        self._fit.add_parameter_constraint(name='a', value=2.0, uncertainty=1.0)
        self._fit.add_parameter_constraint(name='b', value=-1.0, uncertainty=0.5)
        self._fit.add_matrix_parameter_constraint(names=['a', 'b'], values=[2.05, -0.95],
                                                  matrix=[[1.1, 0.1], [0.1, 2.4]])

        self.setup_streams()

    def test_read_from_testfile_stream(self):
        _read_fit = self._testfile_streamreader.read()
        self.assertTrue(isinstance(_read_fit, self.FIT_CLASS))
        self.assertTrue(
            np.allclose(
                self._test_parameters_default,
                _read_fit.parameter_values
            )
        )
        self.assertTrue(
            np.allclose(
                self._test_y_default,
                _read_fit.model
            )
        )

        _read_fit.do_fit()

        self.assertTrue(
            np.allclose(
                self._test_parameters_do_fit,
                _read_fit.parameter_values
            )
        )
        self.assertTrue(
            np.allclose(
                self._test_y_do_fit,
                _read_fit.model
            )
        )

    def test_read_from_testfile_stream_simple(self):
        _read_fit = self._testfile_streamreader_simple.read()
        self.assertTrue(isinstance(_read_fit, self.FIT_CLASS))
        self.assertTrue(
            np.allclose(
                self._test_parameters_default_simple,
                _read_fit.parameter_values
            )
        )
        self.assertTrue(
            np.allclose(
                self._test_y_default_simple,
                _read_fit.model
            )
        )
        
        _read_fit.do_fit()
        
        self.assertTrue(
            np.allclose(
                self._test_parameters_do_fit_simple,
                _read_fit.parameter_values
            )
        )
        self.assertTrue(
            np.allclose(
                self._test_y_do_fit_simple,
                _read_fit.model
            )
        )

    def test_round_trip_with_stringstream(self):
        self._roundtrip_streamwriter.write()
        self._roundtrip_stringstream.seek(0)  # return to beginning
        _read_fit = self._roundtrip_streamreader.read()
        self.assertTrue(isinstance(_read_fit, self.FIT_CLASS))

        self.assertTrue(
            np.allclose(
                self._test_parameters_default,
                _read_fit.parameter_values
            )
        )
        self.assertTrue(
            np.allclose(
                self._test_y_default,
                _read_fit.model
            )
        )
        
        _read_fit.do_fit()

        self.assertTrue(
            np.allclose(
                self._test_parameters_do_fit,
                _read_fit.parameter_values
            )
        )
        self.assertTrue(
            np.allclose(
                self._test_y_do_fit,
                _read_fit.model
            )
        )


TEST_FIT_XY="""
type: xy
cost_function: chi2
dataset:
    type: xy
    x_data: [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0]
    y_data: [ -1.0804945, 0.97336504, 2.75769933, 4.91093935, 6.98511206,
             9.15059627, 10.9665515, 13.06741151, 14.95081026, 16.94404467]
    y_errors:
      - correlation_coefficient: 0.0
        error_value: 0.1
        name: test_y_error
        relative: false
        type: simple
parametric_model:
    type: xy
    x_data: [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0]
    model_function:
        python_code: 'x a b -> a * x + b'
    model_parameters: [1.5, -0.5]
parameter_constraints:
      - type: simple
        index: 0
        value: 2.0
        uncertainty: 1.0
      - type: simple
        name: b
        value: -1.0
        uncertainty: 0.5
      - type: matrix
        names: [a, b]
        values: [2.05, -0.95]
        matrix: [[1.1, 0.1], [0.1, 2.4]]
"""

TEST_FIT_XY_MISSING_KEYWORD="""
type: xy
parametric_model:
    type: xy
    x_data: [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0]
    model_function:
        python_code: |
            def linear_model(x, a, b):
                return a * x + b
    model_parameters: [1.5, -0.5]
"""

TEST_FIT_XY_EXTRA_KEYWORD = TEST_FIT_XY + """
extra_keyword: 3.14
"""

TEST_FIT_XY_SIMPLE="""
x_data: [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0]
y_data: [ -1.0804945, 0.97336504, 2.75769933, 4.91093935, 6.98511206,
        9.15059627, 10.9665515, 13.06741151, 14.95081026, 16.94404467]
model_function: linear
y_errors: 0.1
"""

TEST_FIT_XY_RELATIVE_ERROR="""
x_data: [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0]
y_data: [ -1.0804945, 0.97336504, 2.75769933, 4.91093935, 6.98511206,
        9.15059627, 10.9665515, 13.06741151, 14.95081026, 16.94404467]
model_function: linear
x_errors: 10%
y_errors: [10%, 10%, 10%, 10%, 10%, 10%, 10%, 10%, 10%, 10%]
"""

TEST_FIT_XY_RELATIVE_ERROR_MIXED="""
x_data: [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0]
y_data: [ -1.0804945, 0.97336504, 2.75769933, 4.91093935, 6.98511206,
        9.15059627, 10.9665515, 13.06741151, 14.95081026, 16.94404467]
model_function: linear
x_errors: 10%
y_errors: [0.10804945, 10%, 10%, 0.491093935, 10%, 0.915059627, 10%, 10%, 10%, 1.694404467]
"""

class TestXYFitYamlRepresenter(unittest.TestCase, AbstractTestFitRepresenter):
    FIT_CLASS = XYFit
    TEST_FIT = TEST_FIT_XY
    TEST_FIT_SIMPLE = TEST_FIT_XY_SIMPLE
    TEST_FIT_EXTRA_KEYWORD = TEST_FIT_XY_EXTRA_KEYWORD
    TEST_FIT_MISSING_KEYWORD = TEST_FIT_XY_MISSING_KEYWORD

    @staticmethod
    def linear_model(x, a=1.5, b=-0.5):
        return a * x + b
    
    def setUp(self):
        self._test_x = np.arange(10)
        self._test_parameters = np.array([2.0, -1.0])
        self._test_parameters_default = np.array([1.5, -0.5])
        self._test_parameters_default_simple = np.array([1.0, 1.0])
        self._test_parameters_do_fit = np.array([2.0115580399995032, -1.0889949779758534])
        self._test_parameters_do_fit_simple = np.array([2.0117809129092095, -1.090410559090481])
        self._test_parameters_do_fit_relative = np.array([1.98399075, -1.06814031])
        self._test_y = [-1.0804945, 0.97336504, 2.75769933, 4.91093935, 6.98511206,
                        9.15059627, 10.9665515, 13.06741151, 14.95081026, 16.94404467]
        self._test_y_default = self.linear_model(self._test_x, *self._test_parameters_default)
        self._test_y_default_simple = self.linear_model(self._test_x, *self._test_parameters_default_simple)
        self._test_y_do_fit = [-1.08899498, 0.92256306, 2.9341211, 4.94567914, 6.95723718, 8.96879522, 10.98035326,
                               12.9919113, 15.00346934, 17.01502738]

        self._test_y_do_fit_simple = [-1.09041056, 0.92137035, 2.93315127, 4.94493218, 6.95671309,
                                      8.96849401, 10.98027492, 12.99205583, 15.00383674, 17.01561766]
        self._test_y_do_fit_relative = [
            -1.06814032, 0.91585043, 2.89984117, 4.88383192, 6.86782267,
            8.85181341, 10.83580416, 12.81979491, 14.80378565, 16.7877764]

        self._fit = XYFit(
            xy_data=[self._test_x, self._test_y],
            model_function=TestXYFitYamlRepresenter.linear_model
        )
        self._fit.add_error(axis='y', err_val=0.1)
        self._fit.add_parameter_constraint(name='a', value=2.0, uncertainty=1.0)
        self._fit.add_parameter_constraint(name='b', value=-1.0, uncertainty=0.5)
        self._fit.add_matrix_parameter_constraint(names=['a', 'b'], values=[2.05, -0.95],
                                                  matrix=[[1.1, 0.1], [0.1, 2.4]])

        self.setup_streams()
        self._testfile_stringstream_relative = IOStreamHandle(StringIO(TEST_FIT_XY_RELATIVE_ERROR))
        self._testfile_streamreader_relative = FitYamlReader(self._testfile_stringstream_relative)
        self._testfile_stringstream_mixed = IOStreamHandle(
            StringIO(TEST_FIT_XY_RELATIVE_ERROR_MIXED))
        self._testfile_streamreader_mixed = FitYamlReader(self._testfile_stringstream_mixed)

    def test_read_from_testfile_stream(self):
        _read_fit = self._testfile_streamreader.read()
        self.assertTrue(isinstance(_read_fit, self.FIT_CLASS))
        self.assertTrue(
            np.allclose(
                self._test_parameters_default,
                _read_fit.parameter_values
            )
        )
        self.assertTrue(
            np.allclose(
                self._test_y_default,
                _read_fit.y_model
            )
        )
        
        _read_fit.do_fit()

        self.assertTrue(
            np.allclose(
                self._test_parameters_do_fit,
                _read_fit.parameter_values
            )
        )
        self.assertTrue(
            np.allclose(
                self._test_y_do_fit,
                _read_fit.y_model
            )
        )

    def test_read_from_testfile_stream_simple(self):
        _read_fit = self._testfile_streamreader_simple.read()
        self.assertTrue(isinstance(_read_fit, self.FIT_CLASS))
        self.assertTrue(
            np.allclose(
                self._test_parameters_default_simple,
                _read_fit.parameter_values
            )
        )
        self.assertTrue(
            np.allclose(
                self._test_y_default_simple,
                _read_fit.y_model
            )
        )
        
        _read_fit.do_fit()
        
        self.assertTrue(
            np.allclose(
                self._test_parameters_do_fit_simple,
                _read_fit.parameter_values
            )
        )
        self.assertTrue(
            np.allclose(
                self._test_y_do_fit_simple,
                _read_fit.y_model
            )
        )

    def test_read_from_testfile_stream_relative(self):
        _read_fit = self._testfile_streamreader_relative.read()
        self.assertTrue(isinstance(_read_fit, self.FIT_CLASS))
        self.assertTrue(
            np.allclose(
                self._test_parameters_default_simple,
                _read_fit.parameter_values
            )
        )
        self.assertTrue(
            np.allclose(
                self._test_y_default_simple,
                _read_fit.y_model
            )
        )

        _read_fit.do_fit()

        self.assertTrue(
            np.allclose(
                self._test_parameters_do_fit_relative,
                _read_fit.parameter_values
            )
        )
        self.assertTrue(
            np.allclose(
                self._test_y_do_fit_relative,
                _read_fit.y_model
            )
        )

    def test_read_from_testfile_stream_relative_mixed(self):
        _read_fit = self._testfile_streamreader_mixed.read()
        self.assertTrue(isinstance(_read_fit, self.FIT_CLASS))
        self.assertTrue(
            np.allclose(
                self._test_parameters_default_simple,
                _read_fit.parameter_values
            )
        )
        self.assertTrue(
            np.allclose(
                self._test_y_default_simple,
                _read_fit.y_model
            )
        )

        _read_fit.do_fit()

        print(_read_fit.x_total_error)
        print(_read_fit.y_total_error)

        self.assertTrue(
            np.allclose(
                self._test_parameters_do_fit_relative,
                _read_fit.parameter_values
            )
        )
        self.assertTrue(
            np.allclose(
                self._test_y_do_fit_relative,
                _read_fit.y_model
            )
        )

    def test_round_trip_with_stringstream(self):
        self._roundtrip_streamwriter.write()
        self._roundtrip_stringstream.seek(0)  # return to beginning
        _read_fit = self._roundtrip_streamreader.read()
        self.assertTrue(isinstance(_read_fit, self.FIT_CLASS))

        self.assertTrue(
            np.allclose(
                self._test_parameters_default,
                _read_fit.parameter_values
            )
        )
        self.assertTrue(
            np.allclose(
                self._test_y_default,
                _read_fit.y_model
            )
        )
        
        _read_fit.do_fit()
        
        self.assertTrue(
            np.allclose(
                self._test_parameters_do_fit,
                _read_fit.parameter_values
            )
        )
        self.assertTrue(
            np.allclose(
                self._test_y_do_fit,
                _read_fit.y_model
            )
        )


TEST_FIT_UNBINNED = """
cost_function: nll
dataset:
  data: [1.0130614498386674, -1.1349287030839197, -0.10341784343652169, -1.390768642168136, -0.9744105188262357,
         0.6056117355777544, -0.5504328324999447, 0.526390450221898, -1.1119561843051506, 0.8564964441075679,
         -1.4966544904555332, 0.3533780292764612, -2.354860160605102, -1.611507787899526, -0.32604149834680735 ,
         -1.8506525618128302, 0.5587792631003663, 0.8639576320048902, -1.3134430505718173, 2.530614938202886 ,
         -1.1044403418787425, -0.8438885895226831, -0.25603003463019774, -0.4436650881879283, -0.7249153868101906,
         -0.5553144725434371, 0.9361796588314751, 0.9505148635855938, -0.44558819618432904, -0.6020371626047384,
         -0.9187934153258002, -0.7881656831171765, -0.31596343642852404, -0.6590057926479401, 0.8013842370770446,
         -1.3802275845787508, 0.3702055619566078, 0.6100729933463217, 0.9967097423047069, 0.8694021685126967,
         1.6080865393403492, 0.19305357203727197, -0.8846274588919996, 3.0462083207914286, 0.6329501416274081,
         2.1668232578651456, -0.36652740938218903, 0.30092771240660027, 1.7452901977315924, -0.4388534800294709,
         1.3918081589095852, -0.6437491809609527, -0.510287038125732, 0.9606898464368234, 0.41806431328828525,
         0.5957760238098067, -1.2102742486664348, 0.44097254257312235, 0.3412064563213552, -1.7451310421167718,
         0.2081663284234683, -1.1066854012184062, 0.4121145273973152, 1.948498429389019, 1.4469448685636854,
         0.07981317629602669, 1.411704446184254, -0.4610151277847929, 0.9477286127925121, 0.08899785814182719]
fixed_parameters:
    mu: 0.1
parametric_model:
    model_function:
        python_code: |
            def normal_distribution_pdf(x, mu, sigma):
                return np.exp(-0.5 * ((x - mu) / sigma) ** 2) / np.sqrt(2.0 * np.pi * sigma ** 2)
    data: [1.0130614498386674, -1.1349287030839197, -0.10341784343652169, -1.390768642168136, -0.9744105188262357,
           0.6056117355777544, -0.5504328324999447, 0.526390450221898, -1.1119561843051506, 0.8564964441075679,
           -1.4966544904555332, 0.3533780292764612, -2.354860160605102, -1.611507787899526, -0.32604149834680735 ,
           -1.8506525618128302, 0.5587792631003663, 0.8639576320048902, -1.3134430505718173, 2.530614938202886 ,
           -1.1044403418787425, -0.8438885895226831, -0.25603003463019774, -0.4436650881879283, -0.7249153868101906,
           -0.5553144725434371, 0.9361796588314751, 0.9505148635855938, -0.44558819618432904, -0.6020371626047384,
           -0.9187934153258002, -0.7881656831171765, -0.31596343642852404, -0.6590057926479401, 0.8013842370770446,
           -1.3802275845787508, 0.3702055619566078, 0.6100729933463217, 0.9967097423047069, 0.8694021685126967,
           1.6080865393403492, 0.19305357203727197, -0.8846274588919996, 3.0462083207914286, 0.6329501416274081,
           2.1668232578651456, -0.36652740938218903, 0.30092771240660027, 1.7452901977315924, -0.4388534800294709,
           1.3918081589095852, -0.6437491809609527, -0.510287038125732, 0.9606898464368234, 0.41806431328828525,
           0.5957760238098067, -1.2102742486664348, 0.44097254257312235, 0.3412064563213552, -1.7451310421167718,
           0.2081663284234683, -1.1066854012184062, 0.4121145273973152, 1.948498429389019, 1.4469448685636854,
           0.07981317629602669, 1.411704446184254, -0.4610151277847929, 0.9477286127925121, 0.08899785814182719]
    model_parameters: [1.5, 1.0]
type: unbinned
"""

TEST_FIT_UNBINNED_EXTRA_KEYWORD = TEST_FIT_UNBINNED + """
extra_keyword: 3.14
"""

TEST_FIT_UNBINNED_SIMPLE = """
type: unbinned
data: [1.0130614498386674, -1.1349287030839197, -0.10341784343652169, -1.390768642168136, -0.9744105188262357,
       0.6056117355777544, -0.5504328324999447, 0.526390450221898, -1.1119561843051506, 0.8564964441075679,
       -1.4966544904555332, 0.3533780292764612, -2.354860160605102, -1.611507787899526, -0.32604149834680735 ,
       -1.8506525618128302, 0.5587792631003663, 0.8639576320048902, -1.3134430505718173, 2.530614938202886 ,
       -1.1044403418787425, -0.8438885895226831, -0.25603003463019774, -0.4436650881879283, -0.7249153868101906,
       -0.5553144725434371, 0.9361796588314751, 0.9505148635855938, -0.44558819618432904, -0.6020371626047384,
       0.5957760238098067, -1.2102742486664348, 0.44097254257312235, 0.3412064563213552, -1.7451310421167718,
       0.2081663284234683, -1.1066854012184062, 0.4121145273973152, 1.948498429389019, 1.4469448685636854,
       0.4645792379289075, 0.08603433240025622, -1.2395727054852086, 0.7213492596560613, 0.3314999995659894,
       0.510785216253824, 0.44208577350556294, -0.6908107349334328, 1.2491129589765388, 0.8363348859267793,
       0.07981317629602669, 1.411704446184254, -0.4610151277847929, 0.9477286127925121, 0.08899785814182719]
"""

TEST_FIT_UNBINNED_MISSING_KEYWORD = """
fixed_parameters:
    mu: 0.1
parametric_model:
    python_code: |
        def normal_distribution_pdf(x, mu, sigma):
            return np.exp(-0.5 * ((x - mu) / sigma) ** 2) / np.sqrt(2.0 * np.pi * sigma ** 2)
type: unbinned
"""


class TestUnbinnedFitYamlRepresenter(unittest.TestCase, AbstractTestFitRepresenter):
    FIT_CLASS = UnbinnedFit
    TEST_FIT = TEST_FIT_UNBINNED
    TEST_FIT_SIMPLE = TEST_FIT_UNBINNED_SIMPLE
    TEST_FIT_EXTRA_KEYWORD = TEST_FIT_UNBINNED_EXTRA_KEYWORD
    TEST_FIT_MISSING_KEYWORD = TEST_FIT_UNBINNED_MISSING_KEYWORD

    @staticmethod
    def normal_distribution_pdf(x, mu, sigma):
        return np.exp(-0.5 * ((x - mu) / sigma) ** 2) / np.sqrt(2.0 * np.pi * sigma ** 2)

    def setUp(self):
        self._test_data = [1.0130614498386674, -1.1349287030839197, -0.10341784343652169, -1.390768642168136,
                           -0.9744105188262357, 0.6056117355777544, -0.5504328324999447, 0.526390450221898,
                           -1.1119561843051506, 0.8564964441075679, -1.4966544904555332, 0.3533780292764612,
                           -2.354860160605102, -1.611507787899526, -0.32604149834680735, -1.8506525618128302,
                           0.5587792631003663, 0.8639576320048902, -1.3134430505718173, 2.530614938202886,
                           -1.1044403418787425, -0.8438885895226831, -0.25603003463019774, -0.4436650881879283,
                           -0.7249153868101906, -0.5553144725434371, 0.9361796588314751, 0.9505148635855938,
                           -0.44558819618432904, -0.6020371626047384, -0.9187934153258002, -0.7881656831171765,
                           -0.31596343642852404, -0.6590057926479401, 0.8013842370770446, -1.3802275845787508,
                           0.3702055619566078, 0.6100729933463217, 0.9967097423047069, 0.8694021685126967,
                           1.6080865393403492, 0.19305357203727197, -0.8846274588919996, 3.0462083207914286,
                           0.6329501416274081, 2.1668232578651456, -0.36652740938218903, 0.30092771240660027,
                           1.7452901977315924, -0.4388534800294709, 1.3918081589095852, -0.6437491809609527,
                           -0.510287038125732, 0.9606898464368234, 0.41806431328828525, 0.5957760238098067,
                           -1.2102742486664348, 0.44097254257312235, 0.3412064563213552, -1.7451310421167718,
                           0.2081663284234683, -1.1066854012184062, 0.4121145273973152, 1.948498429389019,
                           1.4469448685636854, 0.07981317629602669, 1.411704446184254, -0.4610151277847929,
                           0.9477286127925121, 0.08899785814182719]
        self._test_parameters = np.array([0.1, -1.0])
        self._test_parameters_default = np.array([0.1, 1.0])
        self._test_parameters_default_simple = np.array([1.0, 1.0])
        self._test_parameters_do_fit = np.array([0.1, 1.09727982])
        self._test_parameters_do_fit_simple = np.array([-0.07060323, 1.02375445])

        self._fit = UnbinnedFit(data=self._test_data, model_density_function='normal_distribution_pdf')
        self._fit.fix_parameter('mu', 0.1)

        self.setup_streams()

    def test_read_from_testfile_stream(self):
        _read_fit = self._testfile_streamreader.read()
        self.assertTrue(isinstance(_read_fit, self.FIT_CLASS))
        self.assertTrue(np.allclose(self._test_parameters_default, _read_fit.parameter_values))
        _read_fit.do_fit()
        self.assertTrue(np.allclose(self._test_parameters_do_fit, _read_fit.parameter_values))

    def test_read_from_testfile_stream_simple(self):
        _read_fit = self._testfile_streamreader_simple.read()

        self.assertTrue(isinstance(_read_fit, self.FIT_CLASS))
        self.assertTrue(np.allclose(
            self._test_parameters_default_simple, _read_fit.parameter_values))
        _read_fit.do_fit()
        self.assertTrue(np.allclose(
            self._test_parameters_do_fit_simple, _read_fit.parameter_values))

    def test_round_trip_with_stringstream(self):
        self._roundtrip_streamwriter.write()
        self._roundtrip_stringstream.seek(0)  # return to beginning
        _read_fit = self._roundtrip_streamreader.read()
        self.assertTrue(isinstance(_read_fit, self.FIT_CLASS))
        self.assertTrue(np.allclose(self._test_parameters_default, _read_fit.parameter_values))
        _read_fit.do_fit()
        self.assertTrue(np.allclose(self._test_parameters_do_fit, _read_fit.parameter_values))



import abc
import inspect
import numpy as np
import six
from collections import OrderedDict

try:
    import sympy as sp
except ModuleNotFoundError:
    sp = None

from .format import ParameterFormatter, ModelFunctionFormatter, latexify_ascii
from ..io.file import FileIOMixin
from ..util import function_library
from ...config import kc


if six.PY2:
    from funcsigs import signature, Signature, Parameter
else:
    from inspect import signature, Signature, Parameter


__all__ = ["ParametricModelBaseMixin", "ModelFunctionBase"]


@six.add_metaclass(abc.ABCMeta)
class ModelFunctionBase(FileIOMixin, object):
    """
    This is a purely abstract class implementing the minimal interface required by all
    model functions.

    In order to be used as a model function, a native Python function must be wrapped
    by an object whose class derives from this base class.
    There is a dedicated :py:class:`ModelFunction` specialization for each type of
    data container.

    This class provides the basic functionality used by all :py:class:`ModelFunction` objects.
    These use introspection (:py:mod:`inspect`) for determining the parameter structure of the
    model function and to ensure the function can be used as a model function (validation).

    """

    FORMATTER_TYPE = ModelFunctionFormatter

    def __init__(self, model_function=function_library.linear_model, independent_argcount=1):
        """
        Construct :py:class:`ModelFunction` object (a wrapper for a native Python function):

        :param model_function: function handle
        :param independent_argcount: The amount of independent variables for this model. The first n variables of the
                                      model function will be treated as independent variables and will not be fitted.
        :type independent_argcount: int
        """
        _custom_defaults = OrderedDict()
        self._name = None

        # determine library function from string specification
        if isinstance(model_function, str):
            self._model_function_handle = function_library.STRING_TO_FUNCTION.get(
                model_function, None)
            if self._model_function_handle is None and "->" in model_function:
                if sp is None:
                    raise ModuleNotFoundError(
                        "SymPy is not installed so it cannot be used to define model functions.")
                _symbol_string, _function_string = model_function.split("->")
                _latex_name = None
                if ":" in _symbol_string:
                    self._name, _symbol_string = _symbol_string.split(":")
                else:
                    self._name = "model"
                _latex_name = latexify_ascii(self._name)
                _symbols = list(sp.symbols(_symbol_string))
                for _i, _symbol_i in enumerate(_symbols):
                    if _i < independent_argcount:
                        continue
                    _symbol_i_str = str(_symbol_i)
                    if "=" in _symbol_i_str:
                        _symbol_i_str, _values_str = _symbol_i_str.split("=")
                        _symbol_i = sp.symbols(_symbol_i_str)
                        _custom_defaults[_symbol_i_str] = float(_values_str)
                        _symbols[_i] = _symbol_i
                _symbolic_function = sp.sympify(_function_string)
                self._model_function_handle = sp.lambdify(_symbols, _symbolic_function)
                _latex_string = sp.latex(_symbolic_function)
                _latex_string = _latex_string.replace(r"{", r"{{")
                _latex_string = _latex_string.replace(r"}", r"}}")
                for _symbol in _symbols:
                    _latex_string = _latex_string.replace(r"{{%s}}" % _symbol, r"{%s}" % _symbol)

                print(_latex_string)
                self._model_function_handle.latex_name = _latex_name
                self._model_function_handle.latex_expression_format_string = _latex_string
            if not self._model_function_handle:
                raise ValueError('Unknown model function: %s' % model_function)
            self._callable = self._model_function_handle

        # special handling of numpy vectorized functions
        elif isinstance(model_function, np.vectorize):
            self._callable = model_function
            self._model_function_handle = model_function.pyfunc

        # handle generic callables
        elif callable(model_function):
            self._model_function_handle = model_function if model_function else self._get_default()
            self._callable = self._model_function_handle

        # raise if not callable
        else:
            raise TypeError("Cannot use {} as model function: object not callable!".format(
                model_function))

        if self._name is None:
            self._name = self._model_function_handle.__name__

        assert int(independent_argcount) >= 0, "The number of independent parameters must be greater than 0"
        self._independent_argcount = int(independent_argcount)
        self._assign_model_function_signature_and_argcount(_custom_defaults)
        self._validate_model_function_raise()
        self._assign_function_formatter()
        self._source_code = None
        super(ModelFunctionBase, self).__init__()

    @classmethod
    def _get_base_class(cls):
        return ModelFunctionBase

    @classmethod
    def _get_object_type_name(cls):
        return 'model_function'

    @classmethod
    def _get_default(cls):
        return function_library.linear_model

    def _assign_model_function_signature_and_argcount(self, custom_defaults={}):
        self._model_function_signature = signature(self._model_function_handle)
        self._model_function_argcount = self._model_function_handle.__code__.co_argcount
        # remove the amount of independent variables from the parameter count
        self._model_function_parcount = self._model_function_argcount - self._independent_argcount
        if custom_defaults:
            self.defaults = [
                custom_defaults.get(_p_name, _p_val)
                for _p_name, _p_val in self.defaults_dict.items()
            ]

    def _validate_model_function_raise(self):
        # evaluate general model function requirements
        for _par in self.signature.parameters.values():
            if _par.kind == _par.VAR_POSITIONAL:
                raise ValueError(
                    "Model function '{}' with variable number of positional "
                    "arguments (*{}) is not supported".format(
                        self._model_function_handle.__name__,
                        _par.name,
                    )
                )
            if _par.kind == _par.VAR_KEYWORD:
                raise ValueError(
                    "Model function '{}' with variable number of keyword "
                    "arguments (**{}) is not supported".format(
                        self._model_function_handle.__name__,
                        _par.name,
                    )
                )
        # require at least one parameter to fit
        if self._model_function_parcount < 1:
            raise ValueError("Model function {0!r} needs at least one parameter besides the "
                             "first {0!s} independent variable(s)!".format(
                             self._model_function_handle, self._independent_argcount))

    def _get_argument_formatters(self):
        return [ParameterFormatter(_arg_name) for _arg_name in self.signature.parameters.keys()]

    def _assign_function_formatter(self):
        self._formatter = self.__class__.FORMATTER_TYPE(
            self.name, arg_formatters=self._get_argument_formatters())
        try:
            _latex_name = self._model_function_handle.latex_name
            if _latex_name is not None:
                self._formatter.latex_name = _latex_name
        except AttributeError:
            pass
        try:
            self._formatter.expression_format_string = \
                self._model_function_handle.expression_format_string
        except AttributeError:
            pass
        try:
            self._formatter.latex_expression_format_string = \
                self._model_function_handle.latex_expression_format_string
        except AttributeError:
            pass

    def __call__(self, *args, **kwargs):
        return self._callable(*args, **kwargs)

    @property
    def name(self):
        """The model function name (a valid Python identifier)"""
        return self._name

    @property
    def func(self):
        """The underlying model function handle"""
        return self._model_function_handle

    @property
    def signature(self):
        """The model function argument specification, as returned by :py:meth:`inspect.signature`"""
        return self._model_function_signature

    @property
    def argcount(self):
        """The number of arguments the model function accepts.
        (including any independent variables which are not parameters)"""
        return self._model_function_argcount

    @property
    def parcount(self):
        """The number of fitting parameters in the model function."""
        return self._model_function_parcount

    @property
    def x_name(self):
        """The name of the independent variable. ``None`` for 0 independent variables."""
        # TODO: don't use case differentiation to maintain compatibility with current fit objects
        if self._independent_argcount == 0:
            return None
        _pars = list(self.signature.parameters.keys())
        if self._independent_argcount == 1:
            return _pars[0]
        return _pars[0:self._independent_argcount]

    @property
    def parameter_names(self):
        """The names of the parameters."""
        return list(self.signature.parameters.keys())[self._independent_argcount:]

    @property
    def formatter(self):
        """The :py:obj:`ModelFunctionFormatter`-derived object for this function"""
        return self._formatter

    @property
    def defaults(self):
        """The default values for model function parameters as a list"""
        return list(self.defaults_dict.values())

    @defaults.setter
    def defaults(self, new_defaults):
        if self.parcount != len(new_defaults):  # first arg is independent variable, but not a parameter
            raise ValueError('Expected %s parameter defaults (1 per parameter), but received %s'
                             % (self.parcount, len(new_defaults)))

        # pad defaults with empties for 'x'
        new_defaults = [Parameter.empty] * (self.argcount - self.parcount) + new_defaults

        # set new signature
        self._model_function_signature = Signature(
            parameters=[
                Parameter(
                    name=_par.name,
                    kind=_par.kind,
                    default=_par_default,
                )
                for _par, _par_default in zip(self.signature.parameters.values(), new_defaults)
            ]
        )

    @property
    def defaults_dict(self):
        """The default values for model function parameters as a dict"""
        _defaults_dict = OrderedDict()
        _x_name = self.x_name  # Actually a list of strings that are used as independent variables
        for _par in self.signature.parameters.values():
            # skip independent variable parameter
            if _x_name is not None and _par.name in _x_name:
                continue
            if _par.default == _par.empty:
                _defaults_dict[_par.name] = kc('core', 'default_initial_parameter_value')
            else:
                _defaults_dict[_par.name] = _par.default

        return _defaults_dict

    @property
    def source_code(self):
        if self._source_code is None:
            return inspect.getsource(self.func)
        return self._source_code


class ParametricModelBaseMixin(object):
    """
    A "mixin" class for representing a parametric model.
    Inheriting from this class in addition to a data container class
    additionally stores a Python function handle referring to the
    model function. The argument structure of this function must
    be compatible with the data container type and it must return
    a numpy array of the same shape as the
    :py:meth:`~kafe2.fit._base.DataContainerBase.data` property of
    the data container.

    This mixin class introduces an additional :py:func:`parameters` property for
    the object, which can be used to obtain and set the values of the parameter

    Derived classes should inherit from :py:class:`ParametricModelBaseMixin` and the
    relevant data container (in that order).
    """
    MODEL_FUNCTION_TYPE = ModelFunctionBase

    def __init__(self, model_func, model_parameters, *args, **kwargs):
        """
        Mixin constructor: sets and initialized the model function.

        :param model_func: handle of Python function (the model function)
        :param model_parameters: iterable of parameter values with which the model function should be initialized
        """
        if isinstance(model_func, self.MODEL_FUNCTION_TYPE):
            self._model_function_object = model_func
        else:
            self._model_function_object = self.MODEL_FUNCTION_TYPE(model_func)
        self.parameters = model_parameters
        super(ParametricModelBaseMixin, self).__init__(*args, **kwargs)
        self.label = "$%s$" % self._model_function_object.formatter.latex_name

    @classmethod
    def _get_base_class(cls):
        return ParametricModelBaseMixin

    @classmethod
    def _get_object_type_name(cls):
        return 'parametric_model'

    @property
    def ndf(self):
        return self.size - self._model_function_object.parcount

    @property
    def parameters(self):
        """Model parameter values"""
        return self._model_parameters

    @parameters.setter
    def parameters(self, parameters):
        """Setter for parameter values"""
        self._model_parameters = parameters

        # flag: recalculate the model values next time they are requested
        self._pm_calculation_stale = True
        self._clear_total_error_cache()  # declared in the container class




import numpy as np
import six
import sys
import matplotlib as mpl
import os

from collections.abc import Iterable
from ...config import kafe2_rc
from ...core.confidence import ConfidenceLevel
from .._base import FitBase
from .._base.format import ScalarFormatter as ScalarBaseFormatter
from matplotlib import pyplot as plt, rcParams
from matplotlib import gridspec as gs
from matplotlib import ticker as plticker
from matplotlib.axes import Axes
from matplotlib import rc_context


__all__ = ["ContoursProfiler"]

_python_version = sys.version_info[0]
_float_template = ".2g" if _python_version == 2 else "#.2g"


def _linear_range_transform(range_, factor, asymmetry=0.0):
    """Return a range that is `factor` larger than `range_`.

    The amount by which the extension is done is the direction of
    each boundary is controlled by a float `asymmetry`, which should
    be between -1 and 1. An `asymmetry` of 0 will extend the range
    symmetrically in the direction of both boundaries. An `asymmetry`
    of -1 (+1) will extend the range in the direction of the first
    (second) boundary, leaving the other boundary unchanged."""

    assert len(range_) == 2
    assert factor > 0.0
    assert -1.0 <= asymmetry <= 1.0

    _m = 0.5 * (range_[1] + range_[0])
    _d = 0.5 * (range_[1] - range_[0])
    _asymm_factor = (1.0 + asymmetry)

    return _m - factor*(2.0 - _asymm_factor) * _d, _m + factor*_asymm_factor*_d


def _maybe_set_tight_layout(figure):
    """Enable 'tight_layout' for a figure if the matplotlib version is high enough."""
    # older versions of matplotlib cause problems with 'tight_layout'
    if not mpl.__version__.startswith('1'):
        figure.set_tight_layout(True)


class SigmaLocator(plticker.Locator):
    """
    Create ticks at evenly spaced offsets from a central value.
    The offsets are integer multiples of a fixed value ('sigma')
    """
    def __init__(self, central_value, sigma):
        self._cval = central_value
        self._sigma = sigma

    def __call__(self):
        """Return the locations of the ticks"""
        _vmin, _vmax = self.axis.get_view_interval()
        return self.tick_values(_vmin, _vmax)

    def tick_values(self, vmin, vmax):
        _n_sigma_dn = int((vmin - self._cval)/self._sigma)
        _n_sigma_up = int((vmax - self._cval)/self._sigma)
        return self.raise_if_exceeds(
            np.arange(_n_sigma_dn, _n_sigma_up+1, 1) * self._sigma + self._cval)


class SigmaFormatter(plticker.Formatter):
    """
    Set the tick labels to indicate the distance to the
    central value, in intervals
    """
    def __init__(self, central_value, sigma):
        """set the tick labels to correspond to sigma"""
        self._cval = central_value
        self._sigma = sigma

    def __call__(self, x, pos=None):
        """Return the format for tick val *x* at position *pos*"""
        # _vmin, _vmax = self.axis.get_data_interval()
        _numeric_label = (x - self._cval)/self._sigma
        _str_label = r"%.2g$\sigma$" % (_numeric_label,)
        return _str_label

    def format_data_short(self, value):
        """Short version of format string for tick"""
        return '{:g}'.format(value)


class ScalarFormatter(plticker.Formatter):
    """Format the tick labels to a specified precision.
    """
    def __init__(self, sigma, n_significant_digits=2):
        """Format the tick label to a specified precision, according to the uncertainty.

        :param float sigma: The uncertainty of the parameter.
        :param int n_significant_digits: Number of significant digits.
        """
        self.formatter = ScalarBaseFormatter(sigma=sigma, n_significant_digits=n_significant_digits)

    def __call__(self, x, pos=None):
        """Return the format for tick val *x* at position *pos*"""
        return self.formatter(x)


class ContoursProfiler(object):
    """
    Object dedicated to *profiling* the cost function in one or two parameters.

    Once a fit has converged and the cost function is at its minimum, the technique of *profiling*
    can be applied to examine the behavior of the cost function around its minimum. Profiling
    involves repeating the cost function minimization with the value of one particular parameter
    being fixed to a series of points in the vicinity of the minimum.
    The function resulting from mapping the parameter values to the cost function value thus
    obtained, is known as a *profile* in that parameter.

    In two dimensions, two parameters can be profiled simultaneously. The resulting profiles are
    functions of two variables and can be visualized in the plane by drawing the 3D *contours*
    along which this function remains constant.

    This object offers a means of calculating both profiles and contours
    """

    _DEFAULT_PLOT_PROFILE_KWARGS = dict(marker='', linewidth=2)
    _DEFAULT_PLOT_PARABOLA_KWARGS = dict(marker='', linewidth=2, linestyle='--')
    _DEFAULT_PLOT_ERROR_SPAN_ON_PROFILE_KWARGS = dict(color='gray', alpha=0.5)
    _DEFAULT_PLOT_MINIMUM_KWARGS = dict(markersize=12, marker='*', linewidth=1.5, linestyle='',
                                        capsize=4, color='green', ecolor='green', elinewidth=1.5)
    _DEFAULT_PLOT_MINIMUM_HVLINES_KWARGS = dict(linewidth=1.0, linestyle='--', color='green', marker='')
    _DEFAULT_PLOT_FILL_CONTOUR_KWARGS = dict(alpha=0.3, linewidth=2)

    def __init__(self, fit_object,
                 profile_points=100, profile_subtract_min=True, profile_bound=2.45,
                 contour_points=100, contour_sigma_values=(1.0, 2.0), contour_smoothing_sigma=0.0,
                 contour_method_kwargs=None):
        """
        Construct a :py:obj:`~kafe2.fit._base.profile.ContoursProfiler` object:

        :param fit_object: fit object for which to calculate profiles/contours
        :type fit_object: :py:obj:`~kafe2.fit._base.FitBase`-derived object
        :param profile_points: number of points at which to sample each profile
        :type profile_points: int
        :param profile_subtract_min: if ``True``, the value of the cost function at the minimum is subtracted from
                                     the profiles, so that only the difference to the minimum is plotted
        :type profile_subtract_min: bool
        :param profile_bound: sample the profile at most this far from the minimum (in sigma)
        :type profile_bound: float
        :param contour_points: number of points at which to sample each contour
        :type contour_points: int
        :param contour_sigma_values: evaluate and show contours for these confidences (in sigma)
        :type contour_sigma_values: iterable of float
        :param contour_smoothing_sigma: apply a smoothing Gaussian filter with this sigma parameter to each contour
                                        (default is ``0.0``, meaning no smoothing)
        :type contour_smoothing_sigma: float
        """
        if not isinstance(fit_object, FitBase):
            raise TypeError("Object %r is not a fit object!" % (fit_object,))

        _contour_confidence_levels = [ConfidenceLevel(n_dimensions=2, sigma=_sigma)
                                      for _sigma in contour_sigma_values]

        self._fit = fit_object
        self._profile_kwargs = dict(points=profile_points, subtract_min=profile_subtract_min, bound=profile_bound)
        self._contour_kwargs = dict(points=contour_points,
                                    confidence_levels=_contour_confidence_levels,
                                    smoothing_sigma=contour_smoothing_sigma,
                                    method_kwargs=contour_method_kwargs)

        self._cost_function_formatted_name = "${}$".format(self._fit._cost_function.formatter.latex_name)
        # FIXME MultiFit does not have an internal _model_function field
        self._parameters_formatted_names = ["${}$".format(pf.latex_name)
                                            for pf in self._fit._get_model_function_parameter_formatters()]

        self._figures = []

    def _make_figure_gs(self, nrows=1, ncols=1):
        _fig = plt.figure(figsize=(8, 8))  # defaults from matplotlibrc

        if mpl.__version__.startswith('1'):
            # old API does not support passing 'figure'
            _gs = gs.GridSpec(nrows=nrows, ncols=ncols)
        else:
            _gs = gs.GridSpec(nrows=nrows, ncols=ncols, figure=_fig)

        self._figures.append(_fig)  # store figure

        return _fig, _gs

    @staticmethod
    def _plot_profile_xy(target_axes, x, y, label):
        _kwargs = ContoursProfiler._DEFAULT_PLOT_PROFILE_KWARGS.copy()
        return target_axes.plot(x, y, label=label, **_kwargs)

    @staticmethod
    def _plot_parabolic_cost(target_axes, x, quad_coeff, x_offset, y_offset, label):
        _kwargs = ContoursProfiler._DEFAULT_PLOT_PARABOLA_KWARGS.copy()
        _y = quad_coeff * (x - x_offset) ** 2 + y_offset
        return target_axes.plot(x, _y, label=label, **_kwargs)

    @staticmethod
    def _plot_error_span_on_profile(target_axes, xmin, xmax, label):
        _kwargs = ContoursProfiler._DEFAULT_PLOT_ERROR_SPAN_ON_PROFILE_KWARGS.copy()

        _err_span_artist = target_axes.axvspan(xmin, xmax, label=label, **_kwargs)

        return _err_span_artist

    @staticmethod
    def _plot_minimum(target_axes, x, y, xerr, yerr, label):
        _kwargs = ContoursProfiler._DEFAULT_PLOT_MINIMUM_KWARGS.copy()

        _min_pt_artists = target_axes.errorbar(x, y, xerr=xerr, yerr=yerr, label=label, **_kwargs)

        # make error bar caps have the same line width as the error bars
        _cap_artists = _min_pt_artists[1]
        for _cap in _cap_artists:
            _cap.set_markeredgewidth(_kwargs.get('elinewidth', 1.5))

        _kwargs = ContoursProfiler._DEFAULT_PLOT_MINIMUM_HVLINES_KWARGS.copy()

        _min_lines_artists = (target_axes.axvline(x, **_kwargs),
                              target_axes.axhline(y, **_kwargs))

        return _min_pt_artists, _min_lines_artists

    @staticmethod
    def _plot_contour_xy(target_axes, contour, label, contour_color):
        _kwargs = ContoursProfiler._DEFAULT_PLOT_FILL_CONTOUR_KWARGS.copy()
        if contour.xy_points is not None:
            return [target_axes.fill(contour.xy_points[0], contour.xy_points[1], label=label, **_kwargs)]
        return [target_axes.contour(contour.grid_x, contour.grid_y, contour.grid_z.T, levels=[0, contour.sigma],
                                    colors="gray", **_kwargs),
                target_axes.contourf(contour.grid_x, contour.grid_y, contour.grid_z.T, levels=[0, contour.sigma],
                                     colors=contour_color, label=label, **_kwargs)]

    # -- public properties

    @property
    def figures(self):
        """The ``matplotlib`` figures managed by this object."""
        return list(self._figures)

    # -- public methods

    # - get numeric profiles/contours

    @staticmethod
    def show(*args, **kwargs):
        """Convenience wrapper for matplotlib.pyplot.show()"""
        plt.show(*args, **kwargs)


    def get_profile(self, parameter, points=None, bound=None, subtract_min=None):
        """
        Calculate and return a profile of the cost function in a parameter.

        :param parameter: name of the parameter to profile (referred to as *x* here).
        :type parameter: str
        :param points: number of equidistant points to use as *x* values
        :type points: int
        :param bound: parameter region for the profile. A single value is interpreted as a sigma
            value for an interval symmetric around the cost function minimum. A tuple of two values
            is interpreted as the bounds of an arbitrary *x* interval to be profiled.
        :type bound: int or tuple of int
        :subtract_min: if the minimal cost function value should be subtracted from the profile
            *y* values.
        :type subtract_min: bool
        :return: two-dimensional array of *x* (parameter) values and *y* (cost function) values
        :rtype: two-dimensional array of float
        """
        if bound is None:
            bound = 2
        if points is None:
            points = self._profile_kwargs['points']
        if subtract_min is None:
            subtract_min = self._profile_kwargs['subtract_min']
        _kwargs = dict(bins=points, bound=bound, args=None, subtract_min=subtract_min)
        self._fit._check_dynamic_error_compatibility()
        return self._fit._fitter.profile(parameter, **_kwargs)  # TODO fix for single fit inside multifit

    def get_contours(self, parameter_1, parameter_2, smoothing_sigma=None):
        """
        Calculate and return a list of contours (one for each requested sigma value).

        :param parameter_1: name of the first contour parameter
        :type parameter_1: str
        :param parameter_2:  name of the second contour parameter
        :type parameter_2: str
        :param smoothing_sigma: sigma for smoothing the contour with a gaussian filter
        :type smoothing_sigma: int or float or complex or Iterable
        :return: list of tuples of the form (sigma, contour)
        :rtype: list of 2-tuples of float and 2d-array
        """
        if smoothing_sigma is None:
            smoothing_sigma = self._contour_kwargs['smoothing_sigma']
        _contours = []
        for _cl_obj in self._contour_kwargs['confidence_levels']:
            _contour_method_kwargs = self._contour_kwargs.get('method_kwargs', dict())
            if _contour_method_kwargs is None:
                _contour_method_kwargs = dict()
            # TODO fix for single fit inside multifit
            _cont = self._fit._fitter.contour(parameter_1, parameter_2, sigma=_cl_obj.sigma,
                                              **_contour_method_kwargs)

            # smooth contours if requested
            if smoothing_sigma > 0 and _cont is not None:
                from scipy.ndimage.filters import gaussian_filter
                _cont[0] = gaussian_filter(_cont[0], smoothing_sigma, mode='wrap')
                _cont[1] = gaussian_filter(_cont[1], smoothing_sigma, mode='wrap')

            _contours.append((_cl_obj, _cont))
        self._fit._check_dynamic_error_compatibility()
        return _contours

    # - plot profiles/contours

    def plot_profile(self, parameter, target_axes=None,
                     show_parabolic=True,
                     show_grid=True,
                     show_legend=True,
                     show_fit_minimum=True,
                     show_error_span=True,
                     show_ticks=True,
                     label_ticks_in_sigma=True,
                     label_fit_minimum=True):
        """
        Plot the profile cost function for a parameter.

        :param parameter: name of the parameter to profile in
        :type parameter: str
        :param target_axes: ``Axes`` object (if ``None``, a new figure is created)
        :type target_axes: ``matplotlib`` ``Axes`` or ``None``
        :param show_parabolic: if ``True``, a parabolic approximation of the profile near the minimum is also drawn
        :type show_parabolic: bool`
        :param show_grid: if ``True``, a grid is drawn
        :type show_grid: bool
        :param show_legend: if ``True``, the legend is displayed
        :param show_fit_minimum: if ``True``, the fit minumum is shown as a marker with error bars
        :type show_fit_minimum: bool
        :type show_legend: bool
        :param show_error_span: if ``True``, the parameter error region is shaded
        :type show_error_span: bool
        :param show_ticks: if ``True``, *x* and *y* ticks are displayed
        :type show_ticks: bool
        :param label_ticks_in_sigma: if ``True``, label ticks are in units of 1 sigma
        :type label_ticks_in_sigma: bool
        :param label_fit_minimum: if ``True``, the parameter value and the 1 sigma error
            will be shown as an annotation
        :type label_fit_minimum: bool

        :return: figure containing the plot result
        :rtype: `matplotlib.figure.Figure`
        """
        with rc_context(kafe2_rc):
            if target_axes is None:
                _fig, _gs = self._make_figure_gs(1, 1)
                _axes = plt.subplot(_gs[0, 0])
            else:
                _axes = target_axes

            _par_val = self._fit.parameter_name_value_dict[parameter]
            _par_id = list(self._fit.parameter_name_value_dict.keys()).index(parameter)
            _par_err = self._fit.parameter_errors[_par_id]
            _cost_function_min = self._fit.cost_function_value
            _par_formatted_name = self._parameters_formatted_names[_par_id]

            _x, _y = self.get_profile(parameter)

            _profile_artist = self._plot_profile_xy(_axes, _x, _y,
                                                    label="profile %s" % (self._cost_function_formatted_name,))

            _y_offset = _cost_function_min if not self._profile_kwargs['subtract_min'] else 0.0

            _parabola_artist = None
            if show_parabolic:
                _parabola_artist = self._plot_parabolic_cost(_axes,
                                                             _x,
                                                             quad_coeff=1. / (_par_err**2),
                                                             x_offset=_par_val,
                                                             y_offset=_y_offset,
                                                             label="parabolic approximation")

            _minimum_artist = None
            if show_fit_minimum:
                _minimum_artist = self._plot_minimum(_axes,
                                                     x=_par_val,
                                                     y=_y_offset,
                                                     xerr=_par_err, yerr=None,
                                                     label="fit minimum")

            if label_fit_minimum:
                sigma_str = r"$\sigma_{{{}}} = {:%s}$" % _float_template
                _axes.annotate(
                    '\n'.join([
                        r"$\langle {}\rangle = {}$".format(
                            _par_formatted_name.strip('$'),
                            ScalarBaseFormatter(_par_err, n_significant_digits=2)(_par_val)
                        ),
                        sigma_str.format(
                            _par_formatted_name.strip('$'),
                            _par_err
                        )]),
                    xy=(_par_val, 1),
                    xycoords=('data', 'axes fraction'),
                    xytext=(0, -10),
                    textcoords='offset points',
                    ha='center',
                    va='top'
                )

            _err_span_artist = None
            if show_error_span:
                _xmin, _xmax = _par_val-_par_err, _par_val+_par_err
                _err_span_artist = self._plot_error_span_on_profile(_axes, xmin=_xmin, xmax=_xmax,
                                                                    label="parameter error")

            _axes.set_xlabel(_par_formatted_name)
            _axes.set_ylabel(self._cost_function_formatted_name if not self._profile_kwargs['subtract_min']
                             else '$\\Delta$%s' % self._cost_function_formatted_name)

            if show_legend:
                _axes.legend(loc='center')

            if show_grid:
                _axes.grid('on')

            if show_ticks:
                _loc_x = SigmaLocator(central_value=_par_val, sigma=_par_err)
                _axes.xaxis.set_major_locator(_loc_x)
                if label_ticks_in_sigma:
                    _form_x = SigmaFormatter(central_value=_par_val, sigma=_par_err)
                    _axes.xaxis.set_major_formatter(_form_x)
                else:
                    _form_x = ScalarFormatter(sigma=_par_err, n_significant_digits=2)
                    _axes.xaxis.set_major_formatter(_form_x)

                _loc_y = plticker.MaxNLocator(5)
                _axes.yaxis.set_major_locator(_loc_y)
            else:
                _axes.set_xticks([])
                _axes.set_yticks([])

            # if own figure created -> set tight layout
            if target_axes is None:
                _maybe_set_tight_layout(_fig)

            return _axes.get_figure()

    def plot_contours(self, parameter_1, parameter_2, target_axes=None,
                      show_grid=True,
                      show_legend=True,
                      show_fit_minimum=True,
                      show_ticks=True,
                      label_ticks_in_sigma=True,
                      naming_convention='sigma'):
        """
        Plot the contour for a parameter pair.

        :param parameter_1: name of the parameter appearing on the *x* axis
        :type parameter_1: str
        :param parameter_2:  name of the parameter appearing on the *y* axis
        :type parameter_2: str
        :param target_axes: ``Axes`` object (if ``None``, a new figure is created)
        :type target_axes: ``matplotlib`` ``Axes` or ``None``
        :param show_grid: if ``True``, a grid is drawn
        :type show_grid: bool
        :param show_legend: if ``True``, the legend is displayed
        :type show_legend: bool
        :param show_fit_minimum: if ``True``, the fit minimum is shown as a marker with error bars
        :type show_fit_minimum: bool
        :param show_ticks: if ``True``, *x* and *y* ticks are displayed
        :type show_ticks: bool
        :param label_ticks_in_sigma: if ``True``, label ticks are in units of 1 sigma
        :type label_ticks_in_sigma: bool
        :param naming_convention: if ``'sigma'`` the contour is labelled in sigma, if ``'cl'`` the contour is labelled
                                  in confidence level
        :type naming_convention: str

        :return: figure containing the plot result
        :rtype: `matplotlib.figure.Figure`
        """
        with rc_context(kafe2_rc):
            if target_axes is None:
                _fig, _gs = self._make_figure_gs(1, 1)
                _axes = plt.subplot(_gs[0, 0])
            else:
                _axes = target_axes

            _par_1_id = list(self._fit.parameter_name_value_dict.keys()).index(parameter_1)
            _par_2_id = list(self._fit.parameter_name_value_dict.keys()).index(parameter_2)
            _par_1_formatted_name = self._parameters_formatted_names[_par_1_id]
            _par_2_formatted_name = self._parameters_formatted_names[_par_2_id]

            if naming_convention.lower() == 'cl':
                _use_cl_in_label = True
            elif naming_convention.lower() == 'sigma':
                _use_cl_in_label = False
            else:
                raise ValueError("Unknown contour naming convention '%s'! "
                                                "Must be one of: ('cl', 'sigma')"
                                                % (naming_convention,))

            _cl_contour_pairs = self.get_contours(parameter_1, parameter_2)

            _contour_artists = []
            for _cl_contour_pair, _prop_cycler in zip(_cl_contour_pairs, rcParams["axes.prop_cycle"]):
                _cl, _contour_xy = _cl_contour_pair
                _artists = []
                if _contour_xy is not None:
                    _label = "%s contour" % (_cl.cl_latex_string if _use_cl_in_label else _cl.sigma_latex_string)
                    _artists = self._plot_contour_xy(_axes, _contour_xy, label=_label,
                                                     contour_color=_prop_cycler["color"])
                _contour_artists += _artists

            _par_1_val = self._fit.parameter_name_value_dict[parameter_1]
            _par_2_val = self._fit.parameter_name_value_dict[parameter_2]
            _par_1_err = self._fit.parameter_errors[_par_1_id]
            _par_2_err = self._fit.parameter_errors[_par_2_id]

            _minimum_artist = None
            if show_fit_minimum:
                _minimum_artist = self._plot_minimum(_axes, x=_par_1_val, y=_par_2_val,
                                                     xerr=_par_1_err, yerr=_par_2_err,
                                                     label="fit minimum")

            _axes.set_xlabel(_par_1_formatted_name)
            _axes.set_ylabel(_par_2_formatted_name)

            if show_legend:
                _axes.legend(loc='best')

            if show_grid:
                _axes.grid('on')

            if show_ticks:
                _loc_x = SigmaLocator(central_value=_par_1_val, sigma=_par_1_err)
                _loc_y = SigmaLocator(central_value=_par_2_val, sigma=_par_2_err)
                _axes.xaxis.set_major_locator(_loc_x)
                _axes.yaxis.set_major_locator(_loc_y)
                if label_ticks_in_sigma:
                    _form_x = SigmaFormatter(central_value=_par_1_val, sigma=_par_1_err)
                    _form_y = SigmaFormatter(central_value=_par_2_val, sigma=_par_2_err)
                    _axes.xaxis.set_major_formatter(_form_x)
                    _axes.yaxis.set_major_formatter(_form_y)
                else:
                    _form_x = ScalarFormatter(sigma=_par_1_err, n_significant_digits=2)
                    _form_y = ScalarFormatter(sigma=_par_2_err, n_significant_digits=2)
                    _axes.xaxis.set_major_formatter(_form_x)
                    _axes.yaxis.set_major_formatter(_form_y)
            else:
                _axes.set_xticks([])
                _axes.set_yticks([])

            # if own figure created -> set tight layout
            if target_axes is None:
                _maybe_set_tight_layout(_fig)

            return _axes.get_figure()

    def plot_profiles_contours_matrix(self,
                                      parameters=None,
                                      show_grid_for=None,
                                      show_ticks_for='all',
                                      show_fit_minimum_for='contours',
                                      show_legend=True,
                                      show_parabolic_profiles=True,
                                      show_error_span_profiles=False,
                                      full_matrix=False,
                                      label_ticks_in_sigma=True,
                                      contour_naming_convention='sigma'):
        """
        Plot all profiles and contours to subplots arranges in a matrix-like fashion.

        :param parameters: parameters for which to display profiles and contours. If ``None``, all parameters.
        :type parameters: list of parameter names or ``None``
        :param show_grid_for: subplots for which to show a grid
        :type show_grid_for: ``None``, ``'profiles'``, ``'contours'`` or ``'all'``
        :param show_ticks_for: subplots for which to show ticks
        :type show_ticks_for: ``None``, ``'profiles'``, ``'contours'`` or ``'all'``
        :param show_fit_minimum_for: subplots for which to show the fit minimum
        :type show_fit_minimum_for: ``None``, ``'profiles'``, ``'contours'`` or ``'all'``
        :param show_legend: if ``True``, the legend is displayed
        :type show_legend: bool
        :param show_parabolic_profiles: if ``True``, a parabolic approximation of the profile near the minimum is also
                                        drawn
        :type show_parabolic_profiles: bool
        :param show_error_span_profiles: if ``True``, the parameter uncertainty region is shaded in the
                                         profile plots
        :type show_error_span_profiles: bool
        :param full_matrix: if ``True``, contour subplots are also shown above the main diagonal
        :type full_matrix: bool
        :param label_ticks_in_sigma: if ``True``, label ticks are in units of 1 sigma
        :type label_ticks_in_sigma: bool
        :param contour_naming_convention: if ``'sigma'`` the contour is labelled in sigma, if ``'cl'`` the contour is
                                          labelled in confidence level
        :type contour_naming_convention: str

        :return: figure containing the plot result
        :rtype: `matplotlib.figure.Figure`
        """
        with rc_context(kafe2_rc):
            _par_names = parameters
            if _par_names is None:
                _par_names = list(self._fit.parameter_name_value_dict.keys())
            else:
                # check if there are any unknown parameters
                _unknown_parameters = set(_par_names) - set(self._fit.parameter_name_value_dict.keys())
                if _unknown_parameters:
                    raise ValueError("Unknown parameters: {}".format(_unknown_parameters))

            # # check if any parameters are fixed and exclude them from the matrix:
            # # TODO: public interface for querying parameter status
            # _free_pars = set(self._fit._fitter._fit_pars)
            # _par_names = [_par for _par in _par_names if _par in _free_pars]

            _npar = len(_par_names)
            _fig, _gs = self._make_figure_gs(_npar, _npar)

            _show_spec_options = ('all', 'profiles', 'contours')

            if show_grid_for is not None and show_grid_for not in _show_spec_options:
                raise ValueError("Unknown specification '%s' for 'show_grid_for'. "
                                 "Expected: one of %r" % (show_grid_for, _show_spec_options))
            if show_ticks_for is not None and show_ticks_for not in _show_spec_options:
                raise ValueError("Unknown specification '%s' for 'show_ticks_for'. "
                                 "Expected: one of %r" % (show_ticks_for, _show_spec_options))
            if show_fit_minimum_for is not None and show_fit_minimum_for not in _show_spec_options:
                raise ValueError("Unknown specification '%s' for 'show_fit_minimum_for'. "
                                 "Expected: one of %r" % (show_fit_minimum_for, _show_spec_options))

            # determine which plot elements to show for each subplot type
            _show_grid_profiles = show_grid_for in ('all', 'profiles')
            _show_grid_contours = show_grid_for in ('all', 'contours')
            _show_ticks_profiles = show_ticks_for in ('all', 'profiles')
            _show_ticks_contours = show_ticks_for in ('all', 'contours')
            _show_minimum_profiles = show_fit_minimum_for in ('all', 'profiles')
            _show_minimum_contours = show_fit_minimum_for in ('all', 'contours')

            _all_legend_handles = tuple()
            _all_legend_labels = tuple()

            # draw profiles to subplots on the diagonal
            _subplots = np.empty((_npar, _npar), dtype=Axes)  # store subplot system in numpy array
            for row in six.moves.range(_npar):
                _axes = _subplots[row, row] = _fig.add_subplot(_gs[row, row])
                self.plot_profile(_par_names[row], target_axes=_axes,
                                  show_parabolic=show_parabolic_profiles,
                                  show_grid=_show_grid_profiles,
                                  show_legend=False,
                                  show_fit_minimum=_show_minimum_profiles,
                                  show_error_span=show_error_span_profiles,
                                  label_ticks_in_sigma=label_ticks_in_sigma,
                                  show_ticks=_show_ticks_profiles)

                if show_legend:
                    _hs, _ls = _axes.get_legend_handles_labels()
                    _all_legend_handles += tuple(_hs)
                    _all_legend_labels += tuple(_ls)

            # draw contours to subplots in the lower (and possibly upper) triangle
            for row in six.moves.range(_npar):
                for col in six.moves.range(row):
                    _axes = _subplots[row, col] = _fig.add_subplot(_gs[row, col])
                    self.plot_contours(_par_names[col], _par_names[row],
                                       target_axes=_axes,
                                       show_grid=_show_grid_contours,
                                       show_legend=False,
                                       show_fit_minimum=_show_minimum_contours,
                                       show_ticks=_show_ticks_contours,
                                       label_ticks_in_sigma=label_ticks_in_sigma,
                                       naming_convention=contour_naming_convention)

                    if show_legend:
                        _hs, _ls = _axes.get_legend_handles_labels()
                        _all_legend_handles += tuple(_hs)
                        _all_legend_labels += tuple(_ls)

                    if full_matrix:
                        _axes = _subplots[col, row] = _fig.add_subplot(_gs[col, row])
                        self.plot_contours(_par_names[row], _par_names[col],
                                           target_axes=_axes,
                                           show_grid=_show_grid_contours,
                                           show_legend=False,
                                           show_fit_minimum=_show_minimum_contours,
                                           show_ticks=_show_ticks_contours,
                                           label_ticks_in_sigma=label_ticks_in_sigma,
                                           naming_convention=contour_naming_convention)

            # post-processing: join column x axes and row y axes (where applicable)
            for i in six.moves.range(_npar):
                _pf_axes = _subplots[i, i]

                _ct_axes_col = np.concatenate([_subplots[:i, i], _subplots[i+1:, i]])
                _ct_axes_col = [_ax for _ax in _ct_axes_col if _ax is not None]

                _pf_axes.get_shared_x_axes().join(_pf_axes, *_ct_axes_col)
                _pf_axes.autoscale()

                _ct_axes_row = np.concatenate([_subplots[i, :i], _subplots[i, i+1:]])
                _ct_axes_row = [_ax for _ax in _ct_axes_row if _ax is not None]
                if len(_ct_axes_row) > 1:
                    _ct_axes_row[0].get_shared_y_axes().join(_ct_axes_row[0], *_ct_axes_row[1:])
                    _ct_axes_row[0].autoscale()

            # more post-processing: adjust axis and tick labels
            for row, _row_plots in enumerate(_subplots):
                for col, _plot in enumerate(_row_plots):
                    # skip empty plots
                    if not _plot:
                        continue

                    _plot.set_xlim(_linear_range_transform(_plot.get_xlim(), factor=1.01))
                    _plot.set_ylim(_linear_range_transform(_plot.get_ylim(), factor=1.01))

                    # adjust y plot range to match x (in sigma)
                    if col != row:
                        _x_lim = _plot.get_xlim()
                        _x_min = self._fit.parameter_values[col]
                        _y_min = self._fit.parameter_values[row]
                        _y_over_x_err_ratio = self._fit.parameter_errors[row]/self._fit.parameter_errors[col]
                        _plot.set_ylim((
                            _y_min + (_x_lim[0] - _x_min)*_y_over_x_err_ratio,
                            _y_min + (_x_lim[1] - _x_min)*_y_over_x_err_ratio
                        ))

                    # hide tick labels except for outer plots
                    if col != 0 and col != row:
                        # not leftmost column or profile -> hide y tick and axis labels
                        _plot.set_ylabel(None)
                        for _label in _plot.get_yticklabels():
                            _label.set_visible(False)
                    if row != _npar - 1:
                        # not bottom row -> hide x tick and axis labels
                        _plot.set_xlabel(None)
                        for _label in _plot.get_xticklabels():
                            _label.set_visible(False)

                    # rotate long x tick labels to avoid overlap
                    if not label_ticks_in_sigma:
                        for _label in _plot.get_xticklabels():
                            _label.set_rotation(90)

            # align x and y axis labels
            try:
                _fig.align_ylabels(_subplots[:, 0])
                _fig.align_xlabels(_subplots[_npar - 1, :])
            except AttributeError:
                # API call not supported by matplotlib version -> ignore
                pass

            if show_legend:
                # suppress multiple entries for the same label
                _hs = []
                _ls = []
                _seen_labels = set()
                for _h, _l in zip(_all_legend_handles, _all_legend_labels):
                    if _l not in _seen_labels:
                        _hs.append(_h)
                        _ls.append(_l)
                        _seen_labels.add(_l)

                if not full_matrix:
                    # attach legend to invisible Axes in top right corner
                    _ax_for_legend = _fig.add_subplot(_gs[0, _npar - 1])
                    _ax_for_legend.set_visible(False)
                    _leg = _fig.legend(
                        _hs, _ls,
                        loc='upper right',
                        borderpad=0,
                        borderaxespad=0
                    )
                    # patch legend bbox to correspond to axes
                    _leg._bbox_to_anchor = _ax_for_legend.bbox
                else:
                    # place legend outside regular Axes in top right corner
                    _ax_for_legend = _subplots[0, _npar - 1]
                    _leg = _ax_for_legend.legend(
                        _hs, _ls,
                        loc='upper left',
                        bbox_to_anchor=(1.1, 0, 1, 1),
                        borderpad=0,
                        borderaxespad=0
                    )

            # old versions cause problems with 'tight_layout'
            _maybe_set_tight_layout(_fig)

            return _fig

    def save(self, fname=None, figures='all', *args, **kwargs):
        """
        Saves the profile figures to files.
        args and kwargs are passed on to matplotlib.Figure.savefig() .

        :param fname: Output file name(s), defaults to profile.png or profile_0.png,
            profile_1.png, ...
        :type fname: None or str or iterable of str.
        :param figures: Which figures to save.
        :type figures: 'all' or int or iterable of int
        """
        if figures == 'all':
            _figure_indices = range(len(self._figures))
        elif isinstance(figures, int):
            _figure_indices = [figures]
        else:
            raise TypeError(f"figures must be either 'all', int or iterable of int, but received type"
                            f"{type(figures)}")
        _file_name_list = fname
        if fname is None:
            if len(self._figures) == 1:
                _file_name_list = ['profile.png']
            else:
                _file_name_list = [f'profile_{_i}.png' for _i in _figure_indices]
        elif isinstance(fname, str):
            if len(_figure_indices) == 1:
                _file_name_list = [fname]
            else:
                name, extension = os.path.splitext(fname)
                _file_name_list = [f'{name}_{_i}{extension}' for _i in _figure_indices]
        elif isinstance(fname, Iterable):
            _file_name_list = fname
        else:
            raise ValueError(
                "fname must be either None, a string, or an iterable, but received type "
                f"{type(fname)}"
            )
        if len(_file_name_list) != len(_figure_indices):
            raise ValueError(
                f'Received {len(_file_name_list)} file names for {len(_figure_indices)} figures.'
                'Must be the same!'
            )
        for _figure_index, _file_name in zip(_figure_indices, _file_name_list):
            self._figures[_figure_index].savefig(_file_name, *args, **kwargs)



import abc
import numpy as np


class AbstractTestFit(abc.ABC):

    MINIMIZER = None

    @abc.abstractmethod
    def setUp(self):
        pass

    @property
    def _fit_result_attributes(self):
        return [
            "cost_function_value", "parameter_names", "parameter_values", "parameter_errors",
            "parameter_cov_mat"
        ]

    def _assert_fit_results_equal(
            self, fit_1, fit_2, rtol=1e-3, atol=1e-6, check_cost_function_value=True,
            fit_2_is_double_fit=False, fit_2_permutation=None):
        for _attr in self._fit_result_attributes:
            if not check_cost_function_value and _attr == "cost_function_value":
                continue
            with self.subTest(attr=_attr):
                _attr_1 = getattr(fit_1, _attr)
                _attr_2 = getattr(fit_2, _attr)
                if fit_2_is_double_fit and _attr_2 is not None:
                    if _attr in ["cost_function_value", "parameter_cov_mat"]:
                        _attr_2 *= 0.5
                    if _attr in ["parameter_errors"] and fit_2.did_fit:
                        _attr_2 *= np.sqrt(2)
                if fit_2_permutation is not None:
                    _attr_2_arr = np.array(_attr_2)
                    if _attr_2_arr.ndim == 1:
                        _attr_2_arr = _attr_2_arr[fit_2_permutation]
                    elif _attr_2_arr.ndim == 2:
                        _attr_2_arr = _attr_2_arr[fit_2_permutation, :][:, fit_2_permutation]
                    if isinstance(_attr_2, np.ndarray):
                        _attr_2 = _attr_2_arr
                    elif isinstance(_attr_2, list):
                        _attr_2 = _attr_2_arr.tolist()
                    elif isinstance(_attr_2, tuple):
                        _attr_2 = tuple(_attr_2_arr)
                self._assert_values_equal(
                    name=_attr, received_value=_attr_2, expected_value=_attr_1,
                    rtol=rtol, atol=atol)

    def _assert_fit_properties(self, fit, properties, rtol=1e-3, atol=1e-6):
        for _attr, _ref_val in properties.items():
            with self.subTest(attr=_attr):
                # retrieve property value
                _attr_val = getattr(fit, _attr)
                self._assert_values_equal(
                    name=_attr, received_value=_attr_val, expected_value=_ref_val,
                    rtol=rtol, atol=atol)

    def _assert_values_equal(self, name, received_value, expected_value, rtol=1e-3, atol=1e-6):
        with self.subTest():
            # check type identical to ref
            self.assertIs(type(received_value), type(expected_value))

            # check value (almost) equal to ref
            try:
                self.assertTrue(
                    (np.all(np.isnan(received_value))
                    and np.all(np.isnan(expected_value))
                    and np.array(received_value).shape == np.array(expected_value).shape)
                    or
                    np.allclose(
                        np.asarray(received_value),
                        np.asarray(expected_value),
                        rtol=rtol,
                        atol=atol,
                    )
                )
            except ValueError:
                # most likely array length mismatch
                print("\nCheck failed: attribute {!r} "
                    "should be:\n\t{}\nand is:\n\t{}".format(
                    name, expected_value, received_value))
                raise
            except TypeError:
                # types contained in array do not support 'allclose'
                try:
                    self.assertEqual(received_value, expected_value)
                except:
                    print("\nCheck failed: attribute {!r} "
                        "should be exactly:\n\t{}\nand is:\n\t{}".format(
                        name, expected_value, received_value))
                    raise
            except:
                _abs_diffs = np.abs(expected_value - received_value)
                _min_abs_diff, _max_abs_diff = np.min(_abs_diffs), np.max(_abs_diffs)

                _rel_diffs = np.abs(received_value / expected_value - 1)
                _min_rel_diff, _max_rel_diff = np.nanmin(_rel_diffs), np.nanmax(_rel_diffs)

                print('_' * 70 +
                    "\nCheck failed: attribute {!r} "
                    "should be approximately:\n\t{}\n"
                    "within:\n\t{!r}\nand is:\n\t{}\n"
                    "(abs diff between: {:g} and {:g})\n"
                    "(rel diff between: {:g} and {:g})\n".format(
                        name, expected_value, dict(rtol=rtol, atol=atol), received_value,
                        _min_abs_diff, _max_abs_diff,
                        _min_rel_diff, _max_rel_diff) + '^' * 70)
                raise

    def _get_test_fits(self):
        pass

    def run_test_for_all_fits(self, ref_prop_dict, call_before_fit=None, fit_names=None, **kwargs):
        for _fit_name, _fit in self._get_test_fits().items():
            # skip non-requested
            if fit_names is not None and _fit_name not in fit_names:
                continue
            with self.subTest(fit=_fit_name):
                # call a user-supplied function
                if call_before_fit:
                    call_before_fit(_fit)

                # test object properties
                self._assert_fit_properties(
                    _fit,
                    ref_prop_dict,
                    **kwargs
                )



from __future__ import print_function
import os


class IOStreamHandle(object):
    """
    Very thin wrapper around a Python `file` object, providing an interface for reading/writing to a stream,
    while also allowing it to be used within a context.
    """

    def __init__(self, buffer):
        """
        Create a handle for an existing I/O buffer (e.g. a stream)
        :param buffer: an Python `file` object (or derived)
        """
        self._buffer = buffer

    def __repr__(self):
        return "{}(buffer={})".format(self.__class__.__name__, self._buffer)

    def __enter__(self):
        # don't open stream here: assume already opened
        return self._buffer  # expose the buffer in a context

    def __exit__(self, *args):
        # don't close stream here: assume need to remain open
        pass

    # -- private methods

    # -- public properties

    @property
    def closed(self):
        if self._buffer is not None:
            return self._buffer.closed
        return True  # inexistent buffer -> handle always closed

    @property
    def mode(self):
        if self._buffer is not None:
            return self._buffer.mode
        return None

    # -- public methods

    def flush(self):
        """Flush the underlying buffer"""
        if self._buffer is None:
            raise IOError("Cannot flush buffer: buffer is None")
        return self._buffer.flush()

    def write(self, content, truncate=False):
        """Write to the underlying buffer"""
        if self._buffer is None:
            raise IOError("Cannot perform write: buffer is None")
        if truncate:
            # TODO: use log system
            print("WARNING: truncation of stream requested, ignored")
        return self._buffer.write(content)

    def read(self, *args, **kwargs):
        """Read from the underlying buffer"""
        if self._buffer is None:
            raise IOError("Cannot perform read:  buffer is None")
        return self._buffer.read(*args, **kwargs)

    def seek(self, offset, *args, **kwargs):
        """Seek a position in the underlying buffer"""
        if self._buffer is None:
            raise IOError("Cannot perform seek:  buffer is None")
        return self._buffer.seek(offset, *args, **kwargs)

    def truncate(self, size=None):
        """Truncate the underlying buffer"""
        if self._buffer is None:
            raise IOError("Cannot truncate:      buffer is None")
        # 'size' must be an integer or missing
        _truncate_args = [size] if size is not None else []
        return self._buffer.truncate(*_truncate_args)


class IOFileHandle(IOStreamHandle):
    """
    This class extends `IOStreamHandle` with functionality for reading from and writing to the filesystem.
    The file can be read or written to either directly or by opening the handle in a context:
    `with IOFileHandle(...) as fh: fh.write(...)`
    """

    _VALID_MODES = ['r', 'a']

    def __init__(self, filename, mode):
        super(IOFileHandle, self).__init__(buffer=None)
        self.filename = filename
        self.mode = mode

    def __repr__(self):
        return "{}(filename={}, mode={})".format(self.__class__.__name__, repr(self.filename), repr(self.mode))

    def __enter__(self):
        """for using an IOFileHandle object in a context"""
        if self._buffer is not None:
            # handle already opened: fail
            raise IOError("Cannot open file '{}': already open!".format(self._filename))
        self._buffer = open(self._filename, self._mode)
        return self._buffer

    def __exit__(self, *args):
        """for using an IOFileHandle object in a context"""
        self._buffer.__exit__()
        self._buffer = None

    # -- private methods

    # -- public properties

    @property
    def mode(self):
        return self._mode

    @mode.setter
    def mode(self, new_mode):
        if new_mode not in self.__class__._VALID_MODES:
            raise ValueError("Unknown IOFileHandle mode '{}': expecting one of {}".format(new_mode, self.__class__._VALID_MODES))
        self._mode = new_mode

    @property
    def filename(self):
        return self._filename

    @filename.setter
    def filename(self, new_filename):
        _abspath = os.path.abspath(new_filename)  # expand relative paths
        if os.path.exists(_abspath) and os.path.isdir(_abspath):
            raise ValueError("Cannot set filename: '{}' is a directory!")
        # TODO: handle inexistent parent directories (?)
        self._filename = new_filename

    # -- public methods

    def write(self, content, truncate=False):
        """Write to the underlying buffer"""
        with self as _fh:
            if truncate:
                _fh.truncate(0)
            _fh.write(content)

    def read(self, *args, **kwargs):
        """Read from the underlying buffer"""
        with self as _fh:
            return _fh.read(*args, **kwargs)


class InputFileHandle(IOFileHandle):
    def __init__(self, filename):
        super(InputFileHandle, self).__init__(filename, mode='r')

    def __repr__(self):
        return "{}(filename={})".format(self.__class__.__name__, repr(self.filename))

    def write(self, content, truncate=False):
        raise IOError("Cannot write to 'InputFileHandle'!")

class OutputFileHandle(IOFileHandle):
    def __init__(self, filename):
        super(OutputFileHandle, self).__init__(filename, mode='a')

    def __repr__(self):
        return "{}(filename={})".format(self.__class__.__name__, repr(self.filename))

    def read(self, *args, **kwargs):
        raise IOError("Cannot read from 'OutputFileHandle'!")


class OutputHandleMultiplexer(IOStreamHandle):
    def __init__(self, *io_handles):
        self._io_handles = io_handles
        self._open_handles = None
        for _fh in self._io_handles:
            if not isinstance(_fh, IOStreamHandle):
                raise ValueError("Cannot initialize OutputHandleMultiplexer: "
                                 "handle object {:r} is of type {}, expected "
                                 "IOStreamHandle!".format(_fh, type(_fh)))

    def __enter__(self):
        self._open_handles = [_fh.__enter__() for _fh in self._io_handles]
        return self

    def __exit__(self, *args):
        for _fh in self._io_handles:
            _fh.__exit__()
        self._open_handles = None

    def write(self, content):
        if self._open_handles is not None:
            for _ofh in self._open_handles:
                _ofh.write(content)
        else:
            with self as _ohm:
                # call a second time from within a context:
                # self._open_handles should no longer be None
                _ohm.write(content)

    def read(self, *args, **kwargs):
        raise IOError("Cannot read from OutputHandleMultiplexer!")

    def flush(self):
        """Flush the underlying buffer"""
        if self._open_handles is None:
            raise IOError("Cannot flush buffers: buffers are not open!")
        for _ofh in self._open_handles:
            _ofh.flush()

    def truncate(self, size=None):
        """Truncate the underlying buffers.
        Note: this fails silently if a truncation operation fails or is not available..."""
        if self._open_handles is None:
            raise IOError("Cannot truncate buffers: buffers are not open!")

        # 'size' must be an integer or missing
        _truncate_args = [size] if size is not None else []
        for _ofh in self._open_handles:
            try:
                _ofh.truncate(*_truncate_args)  # attempt to truncate
            except IOError:
                # for streams that do not support truncation: ignore
                pass


if __name__ == "__main__":
    # TODO: move this to a unit test
    import sys
    # test 1: IOStreamHandle
    print('Instantiating `IOStreamHandle`...')
    _iosh = IOStreamHandle(sys.stdout)
    _iosh.write("1")
    with _iosh as _fh:
        _fh.write("2")
        _fh.write("3\n")

    # test 2: IOFileHandle
    print('Instantiating `OutputFile`')
    _ofh = OutputFileHandle("/tmp/test_IOFileHandle.txt")
    print('Writing...')
    _ofh.write("4", truncate=True)
    with _ofh as _fh:
        _fh.write("5")
        _fh.write("6\n")
    print('...done!')

    print('Trying to read from output stream...')
    try:
        _ofh.read()
    except IOError as e:
        print("...OK! IOError raised:", e)

    _ifh = InputFileHandle("/tmp/test_IOFileHandle.txt")
    print("Reading from file: ")
    print(_ifh.read())

    print('Trying to write to input stream...')
    try:
        _ifh.write("bla")
    except IOError as e:
        print("...OK! IOError raised:", e)

    # test 3: OutputHandleMultiplexer
    print('Instantiating `OutputHandleMultiplexer`')
    _ohm = OutputHandleMultiplexer(_iosh, _ofh)
    print('Writing...')
    _ohm.write("a")
    with _ohm as _fh:
        _fh.write("b")
        _fh.write("c\n")
    print('...done!')

    print("Reading from file: ")
    print(_ifh.read())



import unittest
import numpy as np

from kafe2.fit import HistContainer, HistFit, IndexedContainer, IndexedFit, XYContainer, XYFit


class TestParameterConstraintInHistFit(unittest.TestCase):

    @staticmethod
    def _model_function(x, a, b):
        return a * x + b

    @staticmethod
    def _model_function_antiderivative(x, a, b):
        return 0.5 * a * x ** 2 + b * x

    def _expected_profile_diff(self, res, cov_mat_inv):
        return res.dot(cov_mat_inv).dot(res)

    def _test_consistency(self, constrained_fit, par_cov_mat_inv):
        constrained_fit.do_fit()
        _cost_function = constrained_fit._fitter._fcn_wrapper
        for _i in range(4):
            for _j in range(9):
                _a = self._test_par_values[_i, 0, _j]
                _b = self._test_par_values[_i, 1, _j]
                _profile_constrained = _cost_function(self._test_par_values[_i, 0, _j], self._test_par_values[_i, 1, _j])
                _diff = _profile_constrained - self._profile_no_constraints[_i, _j]
                _expected_profile_diff = self._expected_profile_diff(self._test_par_res[_i, _j], par_cov_mat_inv)
                self.assertTrue(np.abs(_diff - _expected_profile_diff) < 1e-12)

    def setUp(self):
        _bin_edges = [0.0, 1.0, 2.0, 3.0, 4.0, 5.0]
        _data = [
            0.5,
            1.5, 1.5,
            2.5, 2.5, 2.5,
            3.5, 3.5, 3.5, 3.5,
            4.5, 4.5, 4.5, 4.5, 4.5
        ]
        self._means = np.array([3.654, 7.789])
        self._vars = np.array([2.467, 1.543])
        self._cov_mat_uncor = np.array([[self._vars[0], 0.0], [0.0, self._vars[1]]])
        self._cov_mat_uncor_inv = np.linalg.inv(self._cov_mat_uncor)
        self._cov_mat_cor = np.array([[self._vars[0], 0.1], [0.1, self._vars[1]]])
        self._cov_mat_cor_inv = np.linalg.inv(self._cov_mat_cor)
        self._cov_mat_simple_a_inv = np.array([[1.0 / self._vars[0], 0.0], [0.0, 0.0]])
        self._cov_mat_simple_b_inv = np.array([[0.0, 0.0], [0.0, 1.0 / self._vars[1]]])

        self._data_container = HistContainer(n_bins=5, bin_range=(0.0, 5.0), fill_data=_data, dtype=float)
        self._data_container.add_error(err_val=1.0)

        _a_test = np.linspace(start=1, stop=2, num=9, endpoint=True)
        _b_test = np.linspace(start=2, stop=3, num=9, endpoint=True)
        self._test_par_values = np.zeros((4, 2, 9))
        self._test_par_values[0, 0] = _a_test
        self._test_par_values[1, 1] = _b_test
        self._test_par_values[2, 0] = _a_test
        self._test_par_values[2, 1] = _b_test
        self._test_par_values[3, 0] = _a_test
        self._test_par_values[3, 1] = _b_test[::-1]  # reverse order
        self._test_par_res = self._test_par_values - self._means.reshape((1, 2, 1))
        self._test_par_res = np.transpose(self._test_par_res, axes=(0, 2, 1))

        self._fit_no_constraints = HistFit(self._data_container, model_function=self._model_function,
                                           bin_evaluation=self._model_function_antiderivative)
        self._fit_no_constraints.do_fit()
        _cost_function = self._fit_no_constraints._fitter._fcn_wrapper
        self._profile_no_constraints = np.zeros((4, 9))
        for _i in range(4):
            for _j in range(9):
                self._profile_no_constraints[_i, _j] = _cost_function(
                    self._test_par_values[_i, 0, _j],
                    self._test_par_values[_i, 1, _j])

    def test_bad_input_exception(self):
        _fit_with_constraint = HistFit(self._data_container, model_function=self._model_function,
                                       bin_evaluation=self._model_function_antiderivative)
        with self.assertRaises(ValueError):
            _fit_with_constraint.add_parameter_constraint('c', 1.0, 1.0)
        with self.assertRaises(ValueError):
            _fit_with_constraint.add_matrix_parameter_constraint(['a', 'c'], [1.0, 2.0], [[0.2, 0.0], [0.0, 0.1]])
        with self.assertRaises(ValueError):
            _fit_with_constraint.add_matrix_parameter_constraint(['a'], [1.0, 2.0], [[0.2, 0.0], [0.0, 0.1]])

    def test_fit_profile_cov_mat_uncorrelated(self):
        _fit_with_constraint = HistFit(self._data_container, model_function=self._model_function,
                                       bin_evaluation=self._model_function_antiderivative)
        _fit_with_constraint.add_matrix_parameter_constraint(['a', 'b'], self._means, self._cov_mat_uncor)
        self._test_consistency(_fit_with_constraint, self._cov_mat_uncor_inv)
        _fit_with_constraint_alt = HistFit(self._data_container, model_function=self._model_function,
                                           bin_evaluation=self._model_function_antiderivative)
        _fit_with_constraint_alt.add_parameter_constraint('a', self._means[0], np.sqrt(self._vars[0]))
        _fit_with_constraint_alt.add_parameter_constraint('b', self._means[1], np.sqrt(self._vars[1]))
        self._test_consistency(_fit_with_constraint_alt, self._cov_mat_uncor_inv)

    def test_fit_profile_cov_mat_correlated(self):
        _fit_with_constraint = HistFit(self._data_container, model_function=self._model_function,
                                       bin_evaluation=self._model_function_antiderivative)
        _fit_with_constraint.add_matrix_parameter_constraint(['a', 'b'], self._means, self._cov_mat_cor)
        self._test_consistency(_fit_with_constraint, self._cov_mat_cor_inv)

    def test_fit_profile_simple_a(self):
        _fit_with_constraint = HistFit(self._data_container, model_function=self._model_function,
                                       bin_evaluation=self._model_function_antiderivative)
        _fit_with_constraint.add_parameter_constraint('a', self._means[0], np.sqrt(self._vars[0]))
        self._test_consistency(_fit_with_constraint, self._cov_mat_simple_a_inv)

    def test_fit_profile_simple_b(self):
        _fit_with_constraint = HistFit(self._data_container, model_function=self._model_function,
                                       bin_evaluation=self._model_function_antiderivative)
        _fit_with_constraint.add_parameter_constraint('b', self._means[1], np.sqrt(self._vars[1]))
        self._test_consistency(_fit_with_constraint, self._cov_mat_simple_b_inv)


class TestParameterConstraintInIndexedFit(unittest.TestCase):

    def _expected_profile_diff(self, res, cov_mat_inv):
        return res.dot(cov_mat_inv).dot(res)

    def _test_consistency(self, constrained_fit, par_cov_mat_inv):
        constrained_fit.do_fit()
        _cost_function = constrained_fit._fitter._fcn_wrapper
        for _i in range(4):
            for _j in range(9):
                _profile_constrained = _cost_function(self._test_par_values[_i, 0, _j], self._test_par_values[_i, 1, _j])
                _diff = _profile_constrained - self._profile_no_constraints[_i, _j]
                _expected_profile_diff = self._expected_profile_diff(self._test_par_res[_i, _j], par_cov_mat_inv)
                self.assertTrue(np.abs(_diff - _expected_profile_diff) < 1e-12)

    @staticmethod
    def _model(a, b):
        return a * np.arange(5) + b

    def setUp(self):
        _data = [-2.1, 0.2, 1.9, 3.8, 6.1]
        self._means = np.array([3.654, 7.789])
        self._vars = np.array([2.467, 1.543])
        self._cov_mat_uncor = np.array([[self._vars[0], 0.0], [0.0, self._vars[1]]])
        self._cov_mat_uncor_inv = np.linalg.inv(self._cov_mat_uncor)
        self._cov_mat_cor = np.array([[self._vars[0], 0.1], [0.1, self._vars[1]]])
        self._cov_mat_cor_inv = np.linalg.inv(self._cov_mat_cor)
        self._cov_mat_simple_a_inv = np.array([[1.0 / self._vars[0], 0.0], [0.0, 0.0]])
        self._cov_mat_simple_b_inv = np.array([[0.0, 0.0], [0.0, 1.0 / self._vars[1]]])

        self._data_container = IndexedContainer(data=_data)
        self._data_container.add_error(err_val=1.0)

        _a_test = np.linspace(start=0,  stop=4, num=9, endpoint=True)
        _b_test = np.linspace(start=-4, stop=0, num=9, endpoint=True)
        self._test_par_values = np.zeros((4, 2, 9))
        self._test_par_values[0, 0] = _a_test
        self._test_par_values[1, 1] = _b_test
        self._test_par_values[2, 0] = _a_test
        self._test_par_values[2, 1] = _b_test
        self._test_par_values[3, 0] = _a_test
        self._test_par_values[3, 1] = -_b_test
        self._test_par_res = self._test_par_values - self._means.reshape((1, 2, 1))
        self._test_par_res = np.transpose(self._test_par_res, axes=(0, 2, 1))

        self._fit_no_constraints = IndexedFit(self._data_container, model_function=self._model)
        self._fit_no_constraints.do_fit()
        _cost_function = self._fit_no_constraints._fitter._fcn_wrapper
        self._profile_no_constraints = np.zeros((4, 9))
        for _i in range(4):
            for _j in range(9):
                self._profile_no_constraints[_i, _j] = _cost_function(
                    self._test_par_values[_i, 0, _j],
                    self._test_par_values[_i, 1, _j])

    def test_bad_input_exception(self):
        _fit_with_constraint = IndexedFit(self._data_container, model_function=self._model)
        with self.assertRaises(ValueError):
            _fit_with_constraint.add_parameter_constraint('c', 1.0, 1.0)
        with self.assertRaises(ValueError):
            _fit_with_constraint.add_matrix_parameter_constraint(['a', 'c'], [1.0, 2.0], [[0.2, 0.0], [0.0, 0.1]])
        with self.assertRaises(ValueError):
            _fit_with_constraint.add_matrix_parameter_constraint(['a'], [1.0, 2.0], [[0.2, 0.0], [0.0, 0.1]])

    def test_fit_profile_cov_mat_uncorrelated(self):
        _fit_with_constraint = IndexedFit(self._data_container, model_function=self._model)
        _fit_with_constraint.add_matrix_parameter_constraint(['a', 'b'], self._means, self._cov_mat_uncor)
        self._test_consistency(_fit_with_constraint, self._cov_mat_uncor_inv)
        _fit_with_constraint_alt = IndexedFit(self._data_container, model_function=self._model)
        _fit_with_constraint_alt.add_parameter_constraint('a', self._means[0], np.sqrt(self._vars[0]))
        _fit_with_constraint_alt.add_parameter_constraint('b', self._means[1], np.sqrt(self._vars[1]))
        self._test_consistency(_fit_with_constraint_alt, self._cov_mat_uncor_inv)

    def test_fit_profile_cov_mat_correlated(self):
        _fit_with_constraint = IndexedFit(self._data_container, model_function=self._model)
        _fit_with_constraint.add_matrix_parameter_constraint(['a', 'b'], self._means, self._cov_mat_cor)
        self._test_consistency(_fit_with_constraint, self._cov_mat_cor_inv)

    def test_fit_profile_simple_a(self):
        _fit_with_constraint = IndexedFit(self._data_container, model_function=self._model)
        _fit_with_constraint.add_parameter_constraint('a', self._means[0], np.sqrt(self._vars[0]))
        self._test_consistency(_fit_with_constraint, self._cov_mat_simple_a_inv)

    def test_fit_profile_simple_b(self):
        _fit_with_constraint = IndexedFit(self._data_container, model_function=self._model)
        _fit_with_constraint.add_parameter_constraint('b', self._means[1], np.sqrt(self._vars[1]))
        self._test_consistency(_fit_with_constraint, self._cov_mat_simple_b_inv)


class TestParameterConstraintInXYFit(unittest.TestCase):

    def _expected_profile_diff(self, res, cov_mat_inv):
        return res.dot(cov_mat_inv).dot(res)

    def _test_consistency(self, constrained_fit, par_cov_mat_inv):
        constrained_fit.do_fit()
        _cost_function = constrained_fit._fitter._fcn_wrapper
        for _i in range(4):
            for _j in range(9):
                _profile_constrained = _cost_function(self._test_par_values[_i, 0, _j], self._test_par_values[_i, 1, _j])
                _diff = _profile_constrained - self._profile_no_constraints[_i, _j]
                _expected_profile_diff = self._expected_profile_diff(self._test_par_res[_i, _j], par_cov_mat_inv)
                self.assertTrue(np.abs(_diff - _expected_profile_diff) < 1e-12)

    def setUp(self):
        _x = [ 0.0, 1.0, 2.0, 3.0, 4.0]
        _y = [-2.1, 0.2, 1.9, 3.8, 6.1]
        self._means = np.array([3.654, 7.789])
        self._vars = np.array([2.467, 1.543])
        self._cov_mat_uncor = np.array([[self._vars[0], 0.0], [0.0, self._vars[1]]])
        self._cov_mat_uncor_inv = np.linalg.inv(self._cov_mat_uncor)
        self._cov_mat_cor = np.array([[self._vars[0], 0.1], [0.1, self._vars[1]]])
        self._cov_mat_cor_inv = np.linalg.inv(self._cov_mat_cor)
        self._cov_mat_simple_a_inv = np.array([[1.0 / self._vars[0], 0.0], [0.0, 0.0]])
        self._cov_mat_simple_b_inv = np.array([[0.0, 0.0], [0.0, 1.0 / self._vars[1]]])

        self._data_container = XYContainer(x_data=_x, y_data=_y)
        self._data_container.add_error(axis='y', err_val=1.0)

        _a_test = np.linspace(start=0,  stop=4, num=9, endpoint=True)
        _b_test = np.linspace(start=-4, stop=0, num=9, endpoint=True)
        self._test_par_values = np.zeros((4, 2, 9))
        self._test_par_values[0, 0] = _a_test
        self._test_par_values[1, 1] = _b_test
        self._test_par_values[2, 0] = _a_test
        self._test_par_values[2, 1] = _b_test
        self._test_par_values[3, 0] = _a_test
        self._test_par_values[3, 1] = -_b_test
        self._test_par_res = self._test_par_values - self._means.reshape((1, 2, 1))
        self._test_par_res = np.transpose(self._test_par_res, axes=(0, 2, 1))

        self._fit_no_constraints = XYFit(self._data_container)
        self._fit_no_constraints.do_fit()
        _cost_function = self._fit_no_constraints._fitter._fcn_wrapper
        self._profile_no_constraints = np.zeros((4, 9))
        for _i in range(4):
            for _j in range(9):
                self._profile_no_constraints[_i, _j] = _cost_function(
                    self._test_par_values[_i, 0, _j],
                    self._test_par_values[_i, 1, _j])

    def test_bad_input_exception(self):
        _fit_with_constraint = XYFit(self._data_container)
        with self.assertRaises(ValueError):
            _fit_with_constraint.add_parameter_constraint('c', 1.0, 1.0)
        with self.assertRaises(ValueError):
            _fit_with_constraint.add_matrix_parameter_constraint(['a', 'c'], [1.0, 2.0], [[0.2, 0.0], [0.0, 0.1]])
        with self.assertRaises(ValueError):
            _fit_with_constraint.add_matrix_parameter_constraint(['a'], [1.0, 2.0], [[0.2, 0.0], [0.0, 0.1]])

    def test_fit_profile_cov_mat_uncorrelated(self):
        _fit_with_constraint = XYFit(self._data_container)
        _fit_with_constraint.add_matrix_parameter_constraint(['a', 'b'], self._means, self._cov_mat_uncor)
        self._test_consistency(_fit_with_constraint, self._cov_mat_uncor_inv)
        _fit_with_constraint_alt = XYFit(self._data_container)
        _fit_with_constraint_alt.add_parameter_constraint('a', self._means[0], np.sqrt(self._vars[0]))
        _fit_with_constraint_alt.add_parameter_constraint('b', self._means[1], np.sqrt(self._vars[1]))
        self._test_consistency(_fit_with_constraint_alt, self._cov_mat_uncor_inv)

    def test_fit_profile_cov_mat_correlated(self):
        _fit_with_constraint = XYFit(self._data_container)
        _fit_with_constraint.add_matrix_parameter_constraint(['a', 'b'], self._means, self._cov_mat_cor)
        self._test_consistency(_fit_with_constraint, self._cov_mat_cor_inv)

    def test_fit_profile_simple_a(self):
        _fit_with_constraint = XYFit(self._data_container)
        _fit_with_constraint.add_parameter_constraint('a', self._means[0], np.sqrt(self._vars[0]))
        self._test_consistency(_fit_with_constraint, self._cov_mat_simple_a_inv)

    def test_fit_profile_simple_b(self):
        _fit_with_constraint = XYFit(self._data_container)
        _fit_with_constraint.add_parameter_constraint('b', self._means[1], np.sqrt(self._vars[1]))
        self._test_consistency(_fit_with_constraint, self._cov_mat_simple_b_inv)



import abc
import six


__all__ = ["FitEnsembleBase", "FitEnsembleException"]


class FitEnsembleException(Exception):
    pass


@six.add_metaclass(abc.ABCMeta)
class FitEnsembleBase(object):
    """
    Object for generating ensembles of fits to pseudo-data generated according to the
    specified uncertainty model.

    This is a purely abstract class implementing the minimal interface required by all
    types of fit ensembles.
    """

    FIT_TYPE = None



from __future__ import print_function

import contextlib
import numpy as np
import six
import sys

from string import ascii_letters


@contextlib.contextmanager
def numpy_print_options(*args, **kwargs):
    """Context for fine-tuning the printout of numpy arrays"""
    _saved_options = np.get_printoptions()
    # set the specified print options
    np.set_printoptions(*args, **kwargs)
    try:
        # execute the context body
        yield
    finally:
        np.set_printoptions(**_saved_options)


_ALPHANUMERIC = np.array(list(ascii_letters) + list("0123456789"))
def random_alphanumeric(size):
    return "".join(np.random.choice(_ALPHANUMERIC, size=size))


def print_dict_recursive(dct, output_stream=sys.stdout, indent_level=0, indent_width=4, indent_char=' ', filter_func=lambda x: True):
    _max_key_len = max([len(_key) for _key in dct.keys()])
    _key_format = "{{:{0}}}".format(_max_key_len + 5)

    _indent_prefix = indent_char * indent_width * indent_level

    for k, v in six.iteritems(dct):
        if isinstance(v, dict):
            output_stream.write(_indent_prefix + _key_format.format(k+':') + '\n')
            output_stream.flush()
            print_dict_recursive(v, indent_level=indent_level+1,  indent_width=indent_width, indent_char=indent_char, filter_func=filter_func)
        else:
            if filter_func(v):
                _content = "{}".format(v)
                if '\n' in _content:
                    _content = '\n' + _content
                    _content = _content.replace('\n', '\n{}{}'.format(_indent_prefix, indent_char * indent_width))
                output_stream.write(_indent_prefix + _key_format.format(k+':') + _content + '\n')
                output_stream.flush()


def print_dict_as_table(dct, output_stream=sys.stdout, sep='  ', cell_format='%.4g',
                        indent_level=0, indent_width=4, indent_char=' '):

    _indent_prefix = indent_char * indent_width * indent_level

    _column_heads = list(dct.keys())
    _column_arrays = list(dct.values())

    # apply formatting to (numeric) cell contents
    for _icol, _col in enumerate(_column_arrays):
        for _irow, _cell in enumerate(_col):
            try:
                _col[_irow] = cell_format % (_cell,)
            except TypeError:
                pass

    _n_rows = len(_column_arrays[0])
    _col_has_n_rows = [_n_rows == len(_arr) for _arr in _column_arrays]
    if False in _col_has_n_rows:
        _wrong_size_columns = [_ch for _i, _ch in enumerate(_column_heads) if not _col_has_n_rows[_i]]
        raise ValueError("All column sizes must be equal. Offending columns: {}".format(_wrong_size_columns))

    _column_formats = []
    _column_widths = []
    _column_show_heads = []
    for _column_head, _column_array in zip(_column_heads, _column_arrays):
        _column_width = max([len("{}".format(str(_cell).strip())) for _cell in _column_array])
        _show_head = not _column_head.startswith('_')
        _column_show_heads.append(_show_head)
        if _show_head:
            _column_width = max(len(_column_head), _column_width)
        _column_formats.append("{{:<{0}}}".format(_column_width))
        #_column_formats.append("{}")
        _column_widths.append(_column_width)

    # transpose table into row-first order
    _table = []
    for _irow in six.moves.range(_n_rows):
        _table.append([])
    for _icol, _col in enumerate(_column_arrays):
        for _irow, _cell in enumerate(_col):
            #_table[_irow].append(np.asarray(_cell))
            _table[_irow].append(_cell)

    # print table header
    for _ihead, (_column_head, _column_format, _column_show_head) in enumerate(zip(_column_heads, _column_formats, _column_show_heads)):
        _column_heads[_ihead] = _column_format.format(_column_head if _column_show_head else "")

    with numpy_print_options(
        #formatter={'float': cell_format.format},
        precision=6,
        suppress=True,  # suppress printing small floating point values
        linewidth=100000  # prevent line breaks in matrix by setting this to a high value
    ):

        output_stream.write(_indent_prefix + sep.join(_column_heads) + '\n')

        # print head separator
        _head_seps = ["="*_w if _column_show_heads[_icol] else " "*_w for _icol, _w in enumerate(_column_widths)]
        output_stream.write(_indent_prefix + sep.join(_head_seps) + '\n')

        # print table body
        for _irow, _row in enumerate(_table):
            for _icol, (_cell, _column_format) in enumerate(zip(_row, _column_formats)):
                try:
                    _row[_icol] = _column_format.format(str(_cell).strip())
                except TypeError:
                    raise
            output_stream.write(_indent_prefix + sep.join(_row) + '\n')


def get_compact_representation(parameter_names, parameter_values, parameter_errors, parameter_cor_mat,
                               asymmetric_parameter_errors=None, line_prefix='# ', table_format='rst'):
    assert (len(parameter_names) == len(parameter_values) == len(parameter_errors) == parameter_cor_mat.shape[0]
            == parameter_cor_mat.shape[1])
    try:
        import tabulate
        if asymmetric_parameter_errors is None:
            _headers = ['Par name', 'Par val', 'Par err', 'Par cor mat']
        else:
            _headers = ['Par name', 'Par val', 'Par err parabolic', 'Par err down', 'Par err up', 'Par cor mat']
        _data = []

        _reduced_cor_mat = [_cor_mat_row[:_i] for _i, _cor_mat_row in enumerate(parameter_cor_mat.tolist())]
        _cor_mat_row_strs = tabulate.tabulate(tabular_data=_reduced_cor_mat, tablefmt='plain', floatfmt='.2g').split('\n')

        for _i, (_par_name, _par_val, _par_err, _cor_mat_row_str) in enumerate(
                zip(parameter_names, parameter_values, parameter_errors, _cor_mat_row_strs)):
            _row = [_par_name]
            if np.isnan(_par_err) or _par_err == 0.0:  # check if parameter is fixed
                _row.append(_par_val)
                _row.append('fixed')
            else:  # parameter is not fixed, round and add errors
                _sig_fig_err = max(2, -int(np.log10(np.abs(_par_err))) + 1)
                _sig_fig_val = max(_sig_fig_err, -int(np.log10(np.abs(_par_val))) + 2)
                _row.append(round(_par_val, _sig_fig_val))
                _row.append(round(_par_err, _sig_fig_err))
            if asymmetric_parameter_errors is not None:
                for _err in asymmetric_parameter_errors[_i]:  # iterate over up and down error
                    if np.isnan(_err):  # parameter is fixed, no error available
                        _row.append('N/A')
                    else:
                        _sig_err = max(2, -int(np.log10(np.abs(_err))) + 1)
                        _row.append(round(_err, _sig_err))
            _row.append(_cor_mat_row_str)
            _data.append(_row)

        _representation = tabulate.tabulate(tabular_data=_data, headers=_headers, tablefmt=table_format)
        _representation = _representation.replace('\n', '\n' + line_prefix)
        _representation = line_prefix + _representation + '\n'
    except ImportError:
        _representation = '# ERROR: Could not create human-readable table for model parameters because ' \
                          'tabulate is not installed.\n'
    return _representation



import numpy as np

from ..histogram import HistContainer, HistFit
from ..indexed import IndexedContainer, IndexedFit
from ..unbinned import UnbinnedContainer, UnbinnedFit
from ..xy import XYContainer, XYFit

__all__ = ['Fit']


def Fit(data=None, model_function=None, minimizer=None, **kwargs):
    """A convenience wrapper for simple fit creation. For more control over the fit creation use the corresponding Fit
    classes.

    :param data: A :py:obj:`~kafe2.fit._base.DataContainerBase`-derived data container, containing the data
                           used in the fit.
    :type data: IndexedContainer or HistContainer or UnbinnedContainer or XYContainer or list
    :param model_function: The model function used in the fit.
    :param minimizer: The minimizer backend to use for the fit. This can be ``'scipy'``, ``'iminuit'`` or ``'tminuit'``,
        depending on the installed backends.
    :type minimizer: str or None
    :param kwargs: Any further keyword arguments for the according fit types. For more information, refer to their
        respective documentation.
    """
    container_to_fit = {IndexedContainer: IndexedFit,
                        HistContainer: HistFit,
                        UnbinnedContainer: UnbinnedFit,
                        XYContainer: XYFit,
                        list: XYFit,
                        np.ndarray: XYFit}
    try:
        fit_class = container_to_fit[type(data)]
    except KeyError:
        raise TypeError("Unknown or unsupported data container type {}."
                        "Supported types are {}".format(type(data), container_to_fit.keys()))
    # other errors will raise during creation of the fit object
    if model_function is None:
        return fit_class(data, minimizer=minimizer, **kwargs)  # use default model function
    # model function is not always called model_function e.g. histogram fit, don't use keyword!
    return fit_class(data, model_function, minimizer=minimizer, **kwargs)



try:
    import typing  # help IDEs with type-hinting inside docstrings
except ImportError:
    pass
import numpy  # help IDEs with type-hinting inside docstrings
import numpy as np

from scipy.misc import derivative

from .._base import ParametricModelBaseMixin
from .container import XYContainer
from ..util import function_library


__all__ = ['XYParametricModel']


class XYParametricModel(ParametricModelBaseMixin, XYContainer):
    # TODO why is model_function abbreviated as model_func?
    def __init__(self, x_data, model_func=function_library.linear_model,
                 model_parameters=(1.0, 1.0)):
        """Construct an :py:obj:`XYParametricModel` object:

        :param x_data: 1D array containing the *x* values supporting the model
        :type x_data: typing.Collection[float]
        :param model_func: Python function handle of the model function.
        :type model_func: typing.Callable
        :param model_parameters: 1D array containing the parameter values with which the model
            function should be initialized.
        :type model_parameters: typing.Collection[float]
        """
        _x_data_array = np.array(x_data)
        _y_data = model_func(_x_data_array, *model_parameters)
        if np.isscalar(_y_data):
            _y_data = np.ones_like(_x_data_array) * _y_data
        self._pm_calculation_stale = False
        super(XYParametricModel, self).__init__(model_func, model_parameters,
                                                _x_data_array, _y_data)

    # -- private methods

    def _recalculate(self):
        # use parent class setter for 'y'
        XYContainer.y.fset(self, self.eval_model_function())
        self._pm_calculation_stale = False

    # -- public properties

    @property
    def data(self):
        """2D array with shape ``(2, N)`` containing the model predictions.

        :rtype: numpy.ndarray[numpy.ndarray[float]]
        """
        if self._pm_calculation_stale:
            self._recalculate()
        return super(XYParametricModel, self).data

    @data.setter
    def data(self, new_data):
        raise TypeError("Parametric model data cannot be set!")

    @property
    def x(self):
        """1D array containing the *x* support values.

        :rtype: numpy.ndarray[float]
        """
        return super(XYParametricModel, self).x

    @x.setter
    def x(self, new_x):
        # resetting 'x' -> must reset entire data array
        self._data = np.zeros((2, len(new_x)))
        self._data[0] = new_x
        self._pm_calculation_stale = True
        self._clear_total_error_cache()

    @property
    def y(self):
        """1D array containing the *y* values calculated from the *x* support values and the
        current parameters.

        :rtype: numpy.ndarray[float]
        """
        if self._pm_calculation_stale:
            self._recalculate()
        return super(XYParametricModel, self).y

    @y.setter
    def y(self, new_y):
        raise TypeError("Parametric model data cannot be set!")

    # -- public methods

    def eval_model_function(self, x=None, model_parameters=None):
        """Evaluate the model function.

        :param x: 1D array containing the *x* values of the support points. If :py:obj:`None`,
            the model *x* values are used.
        :type x: numpy.ndarray[float]
        :param model_parameters: 1D array containing the values of the model parameters. If
            :py:obj:`None`, the current values are used.
        :type model_parameters: typing.Collection[float] or None
        :return: Values of the model function for the given parameters.
        :rtype: numpy.ndarray[float]
        """
        _x = x if x is not None else self.x
        _pars = model_parameters if model_parameters is not None else self._model_parameters
        _y = self._model_function_object(_x, *_pars)
        if np.isscalar(_y):
            _y = np.ones_like(_x) * _y
        return _y

    def eval_model_function_derivative_by_parameters(self, x=None, model_parameters=None,
                                                     par_dx=None):
        """Evaluate the derivative of the model function with respect to the model parameters.

        :param x: 1D array with length ``N`` containing the *x* values of the support points. If
            :py:obj:`None`, the model *x* values are used.
        :type x: numpy.ndarray[float] or None
        :param model_parameters: 1D array with length ``pars`` containing the values of the model
            parameters. If :py:obj:`None`, the current values are used.
        :type model_parameters: typing.Collection[float] or None
        :param par_dx: 1D array with length ``pars`` containing the numeric differentiation step
            size for each parameter.
        :type par_dx: typing.Collection[float]
        :return: 2D array with shape ``(pars, N)`` containing the values of the model function
            derivatives with respect to the parameters.
        :rtype: numpy.ndarray[numpy.ndarray[float]]
        """
        _x = x if x is not None else self.x
        _pars = model_parameters if model_parameters is not None else self._model_parameters
        _pars = np.asarray(_pars)
        _par_dxs = par_dx if par_dx is not None \
            else 1e-2 * (np.abs(_pars) + 1.0 / (1.0 + np.abs(_pars)))

        _ret = np.zeros((len(_pars), len(_x)))
        for _par_idx, (_par_val, _par_dx) in enumerate(zip(_pars, _par_dxs)):
            def _chipped_func(par):
                _chipped_pars = _pars.copy()
                _chipped_pars[_par_idx] = par
                return self._model_function_object(_x, *_chipped_pars)

            _der_val = np.array(derivative(_chipped_func, _par_val, dx=_par_dx))
            _ret[_par_idx] = _der_val
        return _ret

    def eval_model_function_derivative_by_x(self, x=None, model_parameters=None, dx=None):
        """Evaluate the derivative of the model function with respect to the independent variable.

        :param x: 1D array containing the *x* values of the support points. If :py:obj:`None`,
            the model *x* values are used.
        :type x: numpy.ndarray[float] or None
        :param model_parameters: 1D array containing the values of the model parameters.
            If :py:obj:`None`, the current values are used.
        :type model_parameters: typing.Collection[float] or None
        :param dx: Step size for numeric differentiation.
        :type dx: float or typing.Collection[float]

        :return: 1D array containing the values of the model function derivative for each parameter.
        :rtype: numpy.ndarray[float]
        """
        _x = x if x is not None else self.x
        _pars = model_parameters if model_parameters is not None else self._model_parameters
        _dxs = dx if dx is not None else np.zeros_like(_x)
        _dxs_default = 1e-2 * (np.abs(_x) + 1.0/(1.0+np.abs(_x)))
        _dxs = np.where(_dxs == 0, _dxs_default, _dxs)  # Replace zero values with defaults

        _low = self._model_function_object(_x - _dxs, *_pars)
        _high = self._model_function_object(_x + _dxs, *_pars)
        return 0.5 * (_high - _low) / _dxs



import numpy as np

class ContourFactory():
    
    @staticmethod
    def create_xy_contour(xy_points, sigma):
        xy_points = np.asarray(xy_points)
        _shape = xy_points.shape
        if len(_shape) != 2 or (_shape[0] != 2 and _shape[1] != 2):
            raise ValueError(
                "Explicit contours can only be created from iterables with shape (2,n) or (n,2).")
        if _shape[0] != 2:
            xy_points = xy_points.T
        return Contour(xy_points=xy_points, sigma=sigma)
    
    @staticmethod
    def create_grid_contour(grid_x, grid_y, grid_z, sigma):
        grid_x = np.asarray(grid_x)
        grid_y = np.asarray(grid_y)
        grid_z = np.asarray(grid_z)
        _shape_x = grid_x.shape
        _shape_y = grid_y.shape
        _shape_z = grid_z.shape
        if len(_shape_x) != 1:
            raise ValueError("grid_x needs to be one-dimensional.")
        if len(_shape_y) != 1:
            raise ValueError("grid_y needs to be one-dimensional.")
        if len(_shape_z) != 2:
            raise ValueError("grid_z needs to be two-dimensional.")
        if _shape_x[0] != _shape_z[0]:
            raise ValueError("grid_z needs to be as wide as grid_x is long.")
        if _shape_y[0] != _shape_z[1]:
            raise ValueError("grid_z needs to be as high as grid_y is long.")
        return Contour(grid_x=grid_x, grid_y=grid_y, grid_z=grid_z, sigma=sigma)
    

class Contour(object):
    
    def __init__(self, xy_points=None, grid_x=None, grid_y=None, grid_z=None, sigma=None):
        if sigma is None:
            raise ValueError("sigma must not be None.")
        self._xy_points = xy_points
        self._grid_x = grid_x
        self._grid_y = grid_y
        self._grid_z = grid_z
        self._sigma = sigma

    @property
    def xy_points(self):
        return self._xy_points
    
    @property
    def grid_x(self):
        return self._grid_x
    
    @property
    def grid_y(self):
        return self._grid_y
    
    @property
    def grid_z(self):
        return self._grid_z
    
    @property
    def sigma(self):
        return self._sigma
    
